import functools
from typing import Optional, Sequence

import equinox
import jax
import jax.random
import numpy as np
import optax
import pytest

import haliax as hax

import levanter.tracker
from levanter.callbacks import eval_loss_loop
from levanter.data import AsyncDataset
from levanter.data.mixture import MixtureDataset
from levanter.trainer import Trainer, TrainerConfig
from levanter.utils.jax_utils import key_iterator, local_cpu_mesh
from levanter.utils.py_utils import non_caching_cycle


class Example(equinox.Module):
    x: hax.NamedArray
    y: hax.NamedArray


Block = hax.Axis("Block", 1024)


def platform_of_array(x):
    if isinstance(x, jax.Array):
        return set(d.platform for d in x.devices())
    elif isinstance(x, hax.NamedArray):
        return platform_of_array(x.array)
    else:
        return "cpu"


class LogitDataset(AsyncDataset[Example]):
    def __init__(self, W, noise, x_mask, x_bias, *, key):
        super().__init__()
        self.W = W
        self.noise = noise
        self.x_mask = x_mask
        self.x_bias = x_bias
        self.key = key

        @equinox.filter_jit
        def _make_example(x_block, y_block, offset):
            return Example(x=x_block[Block, offset], y=y_block[Block, offset])

        self._make_example = _make_example

        @functools.lru_cache
        @equinox.filter_jit
        def _gen_block_data(block_id):
            key = jax.random.fold_in(self.key, block_id)
            x_block = hax.random.normal(key, (Block, self.W.axes[0])) * self.x_mask + self.x_bias
            noise = hax.random.normal(key, (Block,)) * self.noise
            y_block = (hax.nn.sigmoid(hax.dot(x_block, self.W, axis=self.W.axes[0]) + noise) > 0.5).astype(float)
            return x_block, y_block

        self._gen_block_data = _gen_block_data

    def _make_block(self, Dim, kk):
        this_key_iter = key_iterator(kk)
        x_block = hax.random.normal(next(this_key_iter), (Block, Dim)) * self.x_mask + self.x_bias
        noise = hax.random.normal(next(this_key_iter), (Block,)) * self.noise
        y_block = (hax.nn.sigmoid(hax.dot(x_block, self.W, axis=Dim) + noise) > 0.5).astype(float)
        return x_block, y_block

    async def async_len(self) -> int:
        raise ValueError("Infinitely long dataset")

    async def final_length_is_known(self) -> bool:
        return False

    def is_finite(self) -> bool:
        return False

    async def current_len(self) -> Optional[int]:
        return None

    async def get_batch(self, indices: Sequence[int]) -> Sequence[Example]:
        blocks = set(i // Block.size for i in indices)

        block_data = {}
        for block_id in blocks:
            x_block, y_block = self._gen_block_data(block_id)
            block_data[block_id] = (x_block, y_block)

        result: list[Example] = []
        indices = np.array(indices, dtype=int)

        for index in indices:
            block_id = index // Block.size
            block_offset = index % Block.size
            x_block, y_block = block_data[block_id]
            result.append(self._make_example(x_block, y_block, block_offset))

        return result


@pytest.mark.slow
def test_estimate_mixture_weights():
    # we create 3 simple logistic regression datasets
    # 1. x is moderately predictive of y (y ~ [0, 0.5, 0.5] x + N(0, noise^2) > 0.5)
    # 2. x is not predictive of y at all, y is highly random (y ~ N(0, 1))
    # 3. x is highly predictive of y, but it's very easy (y = sigmoid([1, 0, 0] x > 0.5)

    Dim = hax.Axis("Dim", 5)
    Batch = hax.Axis("Batch", 32)

    # data loading needs to take place on CPU
    with local_cpu_mesh():
        keys = key_iterator(0)
        W1 = hax.named([0.0, 0.5, 0.5, 0.0, 0.0], (Dim,))
        x1_mask = hax.named([0.0, 1.0, 1.0, 0.0, 0.0], (Dim,))
        W2 = hax.named([0.0, 0.0, 0.0, 0.0, 0.0], (Dim,))
        x2_mask = hax.named([0.0, 0.0, 0.0, 1.0, 1.0], (Dim,))
        W3 = hax.named([1.0, 0.0, 0.0, 0.0, 0.0], (Dim,))
        x3_mask = hax.named([1.0, 0.0, 0.0, 0.0, 0.0], (Dim,))
        x3_bias = hax.named([1.0, 0.0, 0.0, 0.0, 0.0], (Dim,))

        # y = sigmoid(Wx + b + N(0, noise^2)) > 0.5
        ds1 = LogitDataset(W1, 0.1, x1_mask, 0.0, key=next(keys))
        ds2 = LogitDataset(W2, 2.0, x2_mask, 0.0, key=next(keys))
        ds3 = LogitDataset(W3, 0.05, x3_mask, x3_bias, key=next(keys))

    # TODO: remove key as a requirement for models
    def compute_loss_fn(model, example, reduction=hax.mean, reduction_axis=None, key=None):
        del key
        y_pred = model(example.x)
        return hax.nn.binary_cross_entropy_loss(y_pred, example.y, reduction=reduction, reduction_axis=reduction_axis)

    tiny_trainer_config = TrainerConfig(
        num_train_steps=300,
        train_batch_size=Batch.size,
        tracker=(),
        id="kmaklfmaf",
        per_device_parallelism=Batch.size // len(jax.devices()),
    )

    optimizer = optax.adam(1e-2)

    trainer = Trainer(tiny_trainer_config, optimizer, compute_loss_fn)

    def fit_to_dataset(dataset: AsyncDataset):
        initial_model = init_model()
        with trainer:
            state = trainer.initial_state(next(keys), model=initial_model)
            loader = trainer.data_loader(dataset, Batch)
            loader = non_caching_cycle(loader)

            loss = 0.0

            # state = trainer.train(state, loader, run_hooks=False)
            for state in trainer.training_steps(state, loader, run_hooks=False):
                if state.step >= 200:
                    loss += state.loss

            return state.model, (loss / (state.step - 200))

    model_key = next(keys)

    def init_model():
        return hax.nn.Linear.init(
            Dim,
            (),
            use_bias=True,
            key=model_key,
            out_first=True,
        )

    m1, loss1 = fit_to_dataset(ds1)
    m2, loss2 = fit_to_dataset(ds2)
    m3, loss3 = fit_to_dataset(ds3)

    assert loss3 < loss1 < loss2

    datasets = {"d1": ds1, "d2": ds2, "d3": ds3}

    ref_model, ref_loss = fit_to_dataset(
        MixtureDataset(datasets, weights={k: 1 / 3.0 for k in datasets.keys()}, key=next(keys), block_size=2048)
    )

    # let's see the loss on each dataset
    l1_ref = eval_loss_loop(compute_loss_fn, ref_model, trainer.data_loader(ds1, Batch), max_batches=10, name="d1")
    l2_ref = eval_loss_loop(compute_loss_fn, ref_model, trainer.data_loader(ds2, Batch), max_batches=10, name="d2")
    l3_ref = eval_loss_loop(compute_loss_fn, ref_model, trainer.data_loader(ds3, Batch), max_batches=10, name="d3")

    assert l3_ref < l1_ref < l2_ref

    from levanter.doremi import estimate_mixture_weights
    from levanter.tracker import NoopTracker

    with levanter.tracker.current_tracker(NoopTracker()):
        w = estimate_mixture_weights(
            initial_proxy=init_model(),
            ref=ref_model,
            data_sources=datasets,
            trainer_config=tiny_trainer_config,
            key=next(keys),
            loss_fn=compute_loss_fn,
        )

    w1 = w["d1"]
    w2 = w["d2"]
    w3 = w["d3"]

    assert w1 > w3 > w2, (w1, w2, w3)
    assert abs(w1 + w2 + w3 - 1.0) < 1e-3
    assert w2 < 0.05  # the noise distribution should get a very low weight



import equinox as eqx
import jax

from levanter.utils.jax_utils import is_inexact_arrayish


def hvp(f, x, v):
    """Compute the Hessian-vector product of a function."""
    return eqx.filter_jvp(eqx.filter_grad(f), (x,), (v,))[1]


def tree_gaussian_like(key, tree):
    """
    Samples a tree of gaussian noise with the same structure as `tree`, except for leaves which are not inexact arrays,
    for which it returns None
    """
    leaves, structure = jax.tree_util.tree_flatten(tree)
    keys = jax.random.split(key, len(leaves))
    rand_n = lambda x, key: jax.random.normal(key, x.shape) if is_inexact_arrayish(x) else None
    g = jax.tree_util.tree_map(rand_n, leaves, list(keys))
    g = jax.tree_util.tree_unflatten(structure, g)

    return g



import jax
import jax.numpy as jnp
import numpy as np
from chex import assert_trees_all_close

import haliax as hax
from haliax import Axis
from haliax.nn.attention import causal_mask

from levanter.models.longformer import causal_sliding_window_attention


def test_causal_sliding_window_attention_simple():
    # test that we can't attend to something outside of the range
    D = 2
    for L, W in [(10, 5), (15, 5)]:
        Pos = Axis("Pos", L)
        Window = Axis("Window", W)
        Head = Axis("Head", D)

        keys = np.zeros((L, D), dtype=np.float32)
        keys[0, 0] = 100.0  # really want to attend to this
        values = np.zeros((L, D), dtype=np.float32)
        values[0, 1] = 300.0  # check if we did attend

        query = np.ones((L, D), dtype=np.float32)

        query = hax.named(query, (Pos, Head))
        keys = hax.named(keys, (Pos, Head))
        values = hax.named(values, (Pos, Head))

        result = causal_sliding_window_attention(Pos, Window, Head, query, keys, values)
        # we should be able to attend to the previous W positions for each position (including current), so 6-10 can't attend
        # to 0-4 and can't get the 100.0 key
        result = result.rearrange((Pos, Head)).array
        assert_trees_all_close(result[0:W, 1], 300)
        assert_trees_all_close(result[W:, 1], 0)


def test_sliding_window_attention_fancier():
    D = 4
    for L, W in [(2, 1), (2, 2), (4, 2), (10, 5), (15, 5), (16, 2), (15, 3), (10, 10)]:
        Pos = Axis("Pos", L)
        Window = Axis("Window", W)
        Head = Axis("Head", D)

        q_key, k_key, v_key = jax.random.split(jax.random.PRNGKey(0), 3)

        query = hax.random.uniform(q_key, (Pos, Head))
        keys = hax.random.uniform(k_key, (Pos, Head))
        values = hax.random.uniform(v_key, (Pos, Head))

        result = causal_sliding_window_attention(Pos, Window, Head, query, keys, values)
        result = result.rearrange((Pos, Head)).array

        KPos = Axis("KPos", Pos.size)
        keys = keys.rename({Pos: KPos})
        values = values.rename({Pos: KPos})

        diff = hax.arange(Pos).broadcast_axis(KPos) - hax.arange(KPos).broadcast_axis(Pos)
        mask = causal_mask(Pos, KPos) & (diff < Window.size) & (diff >= 0)

        # check that the result is the same as non-blocked attention with the right mask
        expected = hax.nn.attention.dot_product_attention(KPos, Head, query, keys, values, mask=mask)

        expected = expected.rearrange((Pos, Head)).array

        assert_trees_all_close(result, expected, atol=1e-3, rtol=1e-3)


def test_longformer_alibi_bias_pos_invariance():
    D = 1
    W = 32
    H = 1

    L = 4096

    Head = Axis("Head", H)
    Pos = Axis("Pos", L)
    Window = Axis("Window", W)
    Hidden = Axis("Hidden", D)

    # this cycles [31, ..., 0, 31, ..., 0, ...]
    cycle = np.flip(np.arange(W, dtype=np.float32))
    v = np.tile(cycle, L // W).reshape((L, H, D))
    v = hax.named(v, (Pos, Head, Hidden))

    q = hax.ones((Pos, Head, Hidden), dtype=jnp.bfloat16) * 0.001
    k = hax.ones((Pos, Head, Hidden), dtype=jnp.bfloat16) * 0.001

    # bias gets geometrically larger as we go further in the sequence
    # this is especially true if there are a lot of heads
    big_head = hax.Axis("Head", 16)
    # NB: this test doesn't work if you use bfloat16 for biases
    bias = hax.nn.attention.alibi_attention_bias(big_head, Pos, dtype=jnp.float32).slice(big_head, Head, 0)

    attn = causal_sliding_window_attention(Pos, Window, Hidden, q, k, v, bias=bias, attention_dtype=jnp.bfloat16)
    attn = attn.rearrange((Pos, Head, Hidden)).array.reshape(L)

    # final value for each cycle should be the same
    finals = attn[W - 1 :: W]
    assert np.isclose(finals, finals[0], rtol=2e-4).all(), f"finals: {finals}"



import io
import json
import os
import re
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Sized,
    Tuple,
    TypeVar,
)
from urllib.parse import urlparse

import datasets
import fsspec
import numpy as np
import pyarrow.parquet as pq
from google.cloud import storage

from levanter.utils import fsspec_utils

from ..data import AsyncDataset
from ._preprocessor import (
    BatchResult,
    _BatchMapTransform,
    _construct_composite_batch_processor,
    _DatasetTransform,
    _MapTransform,
)
from .utils import batched


if TYPE_CHECKING:
    from .metrics_monitor import MetricsMonitor

T = TypeVar("T")
T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)
U = TypeVar("U")


class ShardedDataSource(Generic[T_co]):
    """
    A ShardedDataset is the main interface for reading data. It's basically a mapping from shard names to iterators,
    with the extra feature that it exposes the ability to skip to a particular row in a shard.

    The difference between a [ShardedDataset][] and a [ShardableDataset][] is that a ShardedDataset
    has a fixed number of shards, and a ShardableDataset `shard` method that can be used to
    split the dataset into multiple shards.
    """

    @property
    def shard_names(self) -> Sequence[str]:
        raise NotImplementedError

    @property
    def num_shards(self) -> int:
        return len(self.shard_names)

    def open_shard(self, shard_name: str) -> Iterator[T_co]:
        return self.open_shard_at_row(shard_name, 0)

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[T_co]:
        raise NotImplementedError

    def __iter__(self):
        """
        Iterate over all data in the dataset, in order.
        """
        for shard_name in self.shard_names:
            for doc in self.open_shard(shard_name):
                yield doc

    def build_or_load_cache(
        self,
        path: str,
        *,
        await_finished: bool = True,
        monitors: Optional[Sequence["MetricsMonitor"]] = None,
    ) -> AsyncDataset[T]:
        """
        Constructs a shard cache version of this dataset using Ray.

        Levanter's preprocessing pipeline offers the following features/guarantees:
        * distributed, sharded preprocessing using Ray
        * deterministic ordering of data
        * interruptible and resumable
        * streaming results (no need to wait for everything to finish)

        Note that this is an experimental API and is subject to change.

        Returns:
            A new AsyncDataset that is backed by the cache.
        """

        source, processor = _construct_composite_batch_processor(self)
        from ..store.cache import build_or_load_cache

        cache = build_or_load_cache(
            path,
            source,
            processor,
            await_finished=await_finished,
            monitors=monitors,
        )
        return cache

    def map(self, fn: Callable[[T_co], U]) -> "ShardedDataSource[U]":
        return _MappedShardedDataSource(self, fn)

    def map_batches(
        self,
        fn: Callable[[list[T_co]], BatchResult],
        batch_size,
        *,
        num_cpus=1,
        num_gpus=0,
        output_exemplar=None,
        **resources,
    ) -> "ShardedDataSource[dict]":
        """
        **Lazily** map a function over batches of data. This is useful for doing things like batching data for a model,
        or for batched preprocessing.

        This function is **lazy**.

        Args:
            fn:  A function that takes a list of data and returns an iterable of results
            batch_size: The batch size to use
            num_cpus: passed to ray
            num_gpus: passed to ray
            **resources: Resources to pass to Ray

        Returns:
            A new ShardedDataset.
        """
        return _BatchMappedShardedDataSource(
            self, fn, batch_size, num_cpus=num_cpus, num_gpus=num_gpus, output_exemplar=output_exemplar, **resources
        )


def gcs_glob(pattern: str) -> list[str]:
    """Glob files in Google Cloud Storage.

    Args:
        pattern: GCS path pattern (gs://bucket/path/*)

    Returns:
        List of matching GCS URLs
    """
    if not pattern.startswith("gs://"):
        # Handle local files
        import glob

        return glob.glob(pattern)

    # Parse bucket and prefix from gs:// URL
    parsed = urlparse(pattern)
    bucket_name = parsed.netloc
    prefix = parsed.path.lstrip("/")

    # Convert glob pattern to regex
    prefix_no_glob = prefix.split("*")[0]
    pattern_as_regex = re.compile(re.escape(prefix).replace("\\*", ".*"))

    # Initialize GCS client
    client = storage.Client()
    bucket = client.bucket(bucket_name)

    # List matching blobs
    matching_urls = []
    for blob in bucket.list_blobs(prefix=prefix_no_glob):
        if pattern_as_regex.match(blob.name):
            matching_urls.append(f"gs://{bucket_name}/{blob.name}")

    return matching_urls


def datasource_from_hf(id: str, *, split, **kwargs) -> ShardedDataSource[dict]:
    """
    Create a ShardedDataset from a HuggingFace dataset. Arguments are passed to load_dataset.
    """
    return WrappedHFDataSource(id, split=split, **kwargs)


def datasource_from_jsonl(urls_or_paths: Sequence[str]) -> ShardedDataSource[dict]:
    return JsonlDataSource(urls_or_paths)


def datasource_from_json(urls_or_paths: Sequence[str]) -> ShardedDataSource[dict]:
    return JsonDataSource(urls_or_paths)


def datasource_from_parquet(urls_or_paths: Sequence[str]) -> ShardedDataSource[dict]:
    return ParquetDataSource(urls_or_paths)


class WrappedHFDataSource(ShardedDataSource[dict]):
    """
    This class is responsible for loading a dataset from HuggingFace Datasets and returning the shards.
    Only (some) IterableDatasets are actually sharded in any meaningful way, so we just return a single shard
    for all other datasets.

    kwargs are passed to load_dataset
    """

    def __init__(self, id, *, split, streaming: bool = True, **kwargs):
        self.id = id
        self.split = split
        self.streaming = streaming
        self.kwargs = kwargs
        self._shard_names = self._compute_shard_names()

    @property
    def shard_names(self) -> Sequence[str]:
        return self._shard_names

    def _compute_shard_names(self):
        dataset = self._load_dataset()
        if isinstance(dataset, datasets.IterableDataset):
            try:
                return [str(i) for i in range(dataset.n_shards)]
            except NotImplementedError:
                return ["data"]
        else:
            return ["data"]

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        dataset = self._load_dataset()
        if isinstance(dataset, datasets.IterableDataset) and shard_name != "data":
            # ex_iterable has a key that gets discarded typically
            shard = map(
                lambda t: t[1],
                dataset._ex_iterable.shard_data_sources(index=int(shard_name), num_shards=dataset.n_shards),
            )
        else:
            shard = dataset

        idx = 0
        for doc in shard:
            if idx >= row:
                yield doc
            idx += 1

    def _load_dataset(self):
        # obnoxiously, the dataset loading stuff doesn't work with ray because of multiprocessing
        # so we have to do this hacky thing where we load the dataset in the worker
        return datasets.load_dataset(self.id, split=self.split, streaming=self.streaming, **self.kwargs)


class TextUrlDataSource(ShardedDataSource[str]):
    """
    Dataset for various text formats.
    """

    def __init__(self, urls, text_key="text"):
        self.urls = urls
        self.text_key = text_key
        self.base_ds = UrlDataSource(urls, columns=[text_key])

    @property
    def shard_names(self) -> Sequence[str]:
        return self.base_ds.shard_names

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[str]:
        url = self.base_ds._shard_name_to_url_mapping[shard_name]
        i = 0
        compression = "infer"
        if url.endswith(".zstd"):  # hacky way to detect zstd
            compression = "zstd"

        format = _sniff_format_for_dataset(url)

        # special case for txt files
        if format == ".txt":
            with fsspec.open(url, "r", compression=compression) as f:
                for line in f:
                    if i >= row:
                        yield line
                    i += 1
        else:
            for doc in self.base_ds.open_shard_at_row(shard_name, row):
                yield doc[self.text_key]


class UrlDataSource(ShardedDataSource[dict]):
    """
    Dataset for various dict-like formats.
    """

    def __init__(self, urls, columns=None):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)
        self.columns = columns

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        url = self._shard_name_to_url_mapping[shard_name]
        i = 0
        compression = "infer"
        if url.endswith(".zstd"):  # hacky way to detect zstd
            compression = "zstd"

        format = _sniff_format_for_dataset(url)
        match format:
            case ".jsonl":
                with fsspec.open(url, "r", compression=compression) as f:
                    # TODO: would be nice if we could seek faster than this. Right now, all we do is skip json parsing
                    # which is not nothing, but not ideal.
                    for line in f:
                        if i >= row:
                            obj = json.loads(line)
                            if self.columns:
                                yield {col: obj[col] for col in self.columns}
                        i += 1
            case ".json":
                with fsspec.open(url, "r", compression=compression) as f:
                    data = json.load(f)
                    for doc in data[row:]:
                        if self.columns:
                            yield {col: doc[col] for col in self.columns}
                        else:
                            yield doc
            case ".parquet":
                with fsspec.open(url, "rb", compression=compression) as f:
                    parquet_file = pq.ParquetFile(f)
                    total_rows = parquet_file.metadata.num_rows
                    if row >= total_rows:
                        return iter([])

                    num_row_groups = parquet_file.metadata.num_row_groups

                    # Compute cumulative row counts
                    row_counts = [parquet_file.metadata.row_group(i).num_rows for i in range(num_row_groups)]
                    cumulative_rows = [0]
                    for count in row_counts:
                        cumulative_rows.append(cumulative_rows[-1] + count)

                    # Find the starting row group and row within it
                    for idx, cum_row in enumerate(cumulative_rows):
                        if cum_row > row:
                            row_group_index = idx - 1
                            start_row_in_group = row - cumulative_rows[row_group_index]
                            break

                    # Read from the starting row group onwards
                    for rg_idx in range(row_group_index, parquet_file.num_row_groups):
                        table = parquet_file.read_row_group(rg_idx, columns=self.columns)
                        if rg_idx == row_group_index:
                            table = table.slice(start_row_in_group)
                        for record in table.to_pylist():
                            yield record
            case _:
                raise ValueError(f"Unknown format {format}")


class AudioTextUrlDataSource(ShardedDataSource[Tuple[np.ndarray, int, str]]):
    """
    Dataset for various audio and text formats.
    """

    def __init__(self, urls, text_key="sentence", audio_key="audio", sampling_rate=16000):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)
        self.text_key = text_key
        self.audio_key = audio_key
        self.sampling_rate = sampling_rate

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    @staticmethod
    def resolve_audio_pointer(audio_pointer, sampling_rate) -> dict[str, Any]:
        import librosa  # noqa F401

        def _load_audio_file(file_name, sampling_rate):
            with fsspec.open(audio_pointer, "rb", compression="infer") as f:
                array, sr = librosa.load(f, sr=sampling_rate)
            return {"array": array, "sampling_rate": sr}

        if isinstance(audio_pointer, dict):
            # These are the 3 ways HuggingFace stores audio in it's Audio type
            # https://huggingface.co/docs/datasets/v2.5.1/en/about_dataset_features#the-audio-type
            if "array" in audio_pointer and "sampling_rate" in audio_pointer:
                audio = audio_pointer
            elif "bytes" in audio_pointer:
                array, sr = librosa.load(io.BytesIO(audio_pointer["bytes"]), sr=sampling_rate)
                audio = {"array": array, "sampling_rate": sr}
            elif "path" in audio_pointer:
                audio = _load_audio_file(audio_pointer["path"], sampling_rate)
            else:
                raise ValueError(f"Unsupported audio format {audio_pointer}")
        elif isinstance(audio_pointer, str):
            # This supports filename pointers to arbitrary audio types
            audio = _load_audio_file(audio_pointer, sampling_rate)
        else:
            raise ValueError(f"Unsupported audio format {audio_pointer}")
        return audio

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[Tuple[np.ndarray, int, str]]:
        url = self._shard_name_to_url_mapping[shard_name]
        i = 0
        with fsspec.open(url, "r", compression="infer") as f:
            format = _sniff_format_for_dataset(url)
            match format:
                case ".jsonl":
                    # TODO: would be nice if we could seek faster than this. Right now, all we do is skip json parsing
                    # which is not nothing, but not ideal.
                    for line in f:
                        if i >= row:
                            mat_json = json.loads(line)
                            audio_pointer = mat_json[self.audio_key]
                            audio = AudioTextUrlDataSource.resolve_audio_pointer(audio_pointer, self.sampling_rate)
                            yield (audio["array"], audio["sampling_rate"], mat_json[self.text_key])
                        i += 1
                case ".json":
                    data = json.load(f)
                    for doc in data[row:]:
                        audio_pointer = doc[self.audio_key]
                        audio = AudioTextUrlDataSource.resolve_audio_pointer(audio_pointer, self.sampling_rate)
                        yield (audio["array"], audio["sampling_rate"], doc[self.text_key])
                case _:
                    raise ValueError(f"Unknown format {format}")


def _sniff_format_for_dataset(url):
    good_formats = [".jsonl", ".txt", ".json", ".parquet"]
    format_from_url = None
    # try both with and without compression (could be gz, bz2, etc, so look at the "first" extension)
    extensions = [os.path.splitext(url)[1], os.path.splitext(os.path.splitext(url)[0])[1]]
    for format in good_formats:
        if any(ext == format for ext in extensions):
            format_from_url = format
            break

    if format_from_url is None:
        raise ValueError(f"Unknown format for {url}")

    if format_from_url == ".json":
        # unfortunately, HF (and others) will use "json" for jsonl files,
        # so we have to do some extra work to distinguish them.
        # Choices are
        # 1. look at the first 2 chars, if the first is "[", then it's probably json.
        #    If it's "{\n", also json. If it's { something else", then it's probably jsonl
        # 2. look at the first line. If it's valid json, then it's probably jsonl, unless there's only one line.
        #
        # (You can't actually distinguish between jsonl and json in a file with one line,
        #  which we'll just declare to be json and not jsonl, since that seems more likely)
        # (1) is cheating a bit, but it's fast and works in most cases we care about. (2) is more robust, but slower.
        with fsspec.open(url, "r", compression="infer") as f:
            first_two = f.read(2)

            if first_two[0] == "[" or first_two == "{\n" or first_two == "{\r":
                return ".json"
            elif first_two[0] == "{":
                return ".jsonl"

            # this is (much) heavier. This is particularly slow if we're dealing with packed/non-prettified json
            # since we're parsing the whole file.
            first_line = first_two + f.readline()
            try:
                json.loads(first_line)
                format_from_url = ".jsonl"
            except json.JSONDecodeError:
                return format_from_url

            if not f.readline():
                # only one line
                format_from_url = ".json"

    return format_from_url


class JsonlDataSource(ShardedDataSource[dict]):
    def __init__(self, urls):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        url = self._shard_name_to_url_mapping[shard_name]
        i = 0
        with fsspec.open(url, "r", compression="infer") as f:
            # TODO: would be nice if we could seek faster than this. Right now, all we do is skip json parsing
            # which is not nothing, but not ideal.
            for line in f:
                if i >= row:
                    yield json.loads(line)
                i += 1


class TextDataSource(ShardedDataSource[dict]):
    def __init__(self, urls):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        url = self._shard_name_to_url_mapping[shard_name]
        i = 0
        with fsspec.open(url, "r", compression="infer") as f:
            for line in f:
                if i >= row:
                    yield line
                i += 1


class JsonDataSource(ShardedDataSource[dict]):
    def __init__(self, urls):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        url = self._shard_name_to_url_mapping[shard_name]
        with fsspec.open(url, "r", compression="infer") as f:
            # TODO: would be nice if we could seek faster than this. Can't even skip json parsing
            data = json.load(f)
            return iter(data[row:])


class ParquetDataSource(ShardedDataSource[dict]):
    def __init__(self, urls):
        self.urls = urls
        self._shard_name_to_url_mapping = _mk_shard_name_mapping(urls)

    @property
    def shard_names(self) -> Sequence[str]:
        return list(self._shard_name_to_url_mapping.keys())

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[dict]:
        url = self._shard_name_to_url_mapping[shard_name]
        with fsspec.open(url, "rb", compression="infer") as f:
            parquet_file = pq.ParquetFile(f)
            total_rows = parquet_file.metadata.num_rows
            if row >= total_rows:
                return iter([])

            num_row_groups = parquet_file.metadata.num_row_groups

            # Compute cumulative row counts
            row_counts = [parquet_file.metadata.row_group(i).num_rows for i in range(num_row_groups)]
            cumulative_rows = [0]
            for count in row_counts:
                cumulative_rows.append(cumulative_rows[-1] + count)

            # find starting row group and also find the row within it
            for idx, cum_row in enumerate(cumulative_rows):
                if cum_row > row:
                    row_group_index = idx - 1
                    start_row_in_group = row - cumulative_rows[row_group_index]
                    break

            # read from the starting row group onwards
            for rg_idx in range(row_group_index, parquet_file.num_row_groups):
                table = parquet_file.read_row_group(rg_idx)

                # if we're in the row group we want, slice the table at/from the row we want
                if rg_idx == row_group_index:
                    table = table.slice(start_row_in_group)

                yield from table.to_pylist()


def _mk_shard_name_mapping(urls):
    _shard_name_to_url_mapping = {}
    # remove common prefix
    if len(urls) == 1:
        common_prefix = os.path.dirname(urls[0])
    else:
        common_prefix = os.path.commonprefix(urls)

    missing_urls: List[str] = []

    for url in urls:
        if not fsspec_utils.exists(url):
            missing_urls.append(url)
            continue
        # escape the url for the shard name
        shard_name = url
        if common_prefix:
            shard_name = url[len(common_prefix) :]
            if shard_name.startswith("/"):
                shard_name = shard_name[1:]

        shard_name = shard_name.replace(".", "_")
        if shard_name in _shard_name_to_url_mapping:
            raise ValueError(f"Duplicate shard name {shard_name}")
        _shard_name_to_url_mapping[shard_name] = url

    if missing_urls:
        # format nicely
        missing_urls_str = "\n  - ".join(missing_urls)
        raise FileNotFoundError(f"Could not find the following urls:\n  - {missing_urls_str}")

    return _shard_name_to_url_mapping


class _TransformedDataset:
    source: ShardedDataSource
    _transform: _DatasetTransform


class _MappedShardedDataSource(ShardedDataSource[T], _TransformedDataset):
    def __init__(self, source: ShardedDataSource[T_co], fn: Callable[[T_co], T]):
        self.source = source
        self.fn = fn
        self._transform = _MapTransform(fn)

    @property
    def shard_names(self) -> Sequence[str]:
        return self.source.shard_names

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[T]:
        for doc in self.source.open_shard_at_row(shard_name, row):
            yield self.fn(doc)


class _BatchMappedShardedDataSource(ShardedDataSource[T], _TransformedDataset):
    def __init__(
        self,
        source: ShardedDataSource[T_co],
        fn: Callable[[list[T_co]], Iterable[U]],
        batch_size,
        num_cpus=1,
        num_gpus=0,
        output_exemplar=None,
        **resources,
    ):
        self.source = source
        self._transform = _BatchMapTransform(
            fn, batch_size, num_cpus, num_gpus, resources, output_exemplar=output_exemplar
        )

    @property
    def shard_names(self) -> Sequence[str]:
        return self.source.shard_names

    def open_shard_at_row(self, shard_name: str, row: int) -> Iterator[T]:
        warnings.warn("This is not the best way to use batched preprocessing. Use build_cache instead.")
        # this one is tricky because we have to do batching ourselves and there's no guarantee that input and output
        # batch sizes are the same
        i = 0
        shard_iter = self.source.open_shard_at_row(shard_name, row)
        for batch in batched(shard_iter, self._transform.batch_size):  # type: ignore
            result = self._transform.fn(batch)  # type: ignore
            if isinstance(result, Sized) and len(result) + i < row:
                i += len(result)
                continue

            for doc in result:
                if i >= row:
                    yield doc
                i += 1



import functools
import logging
import time
from collections import defaultdict
from typing import AsyncIterator, Callable, Iterable, Iterator, Optional, Tuple, TypeVar

import jax
from jax import Array
from jax import numpy as jnp
from jax import tree_util as jtu
from jax.experimental import multihost_utils
from jax.sharding import Mesh, PartitionSpec
from jaxtyping import PyTree

import haliax as hax
from haliax import is_named_array
from haliax._src.util import index_where
from haliax.partitioning import ResourceMapping

from levanter.data.dataset import AsyncDataset
from levanter.data.utils import batched
from levanter.shapes import NamedShapeSpec, ShapeSpec, to_raw_shape
from levanter.utils.background_iterable import BackgroundIterator
from levanter.utils.jax_utils import local_cpu_mesh
from levanter.utils.thread_utils import AsyncIteratorWrapper, blocking_wait


Ex = TypeVar("Ex")

_TensorSliceIndex = tuple[slice, ...]
logger = logging.getLogger(__name__)


class DataLoader(Iterable[Ex]):
    def __init__(
        self,
        Batch: hax.Axis,
        data: AsyncDataset[Ex],
        max_buffered_batches: Optional[int],
        mesh: Mesh,
        axis_resources: Optional[ResourceMapping],
        # this is set heuristically for the typical tokenseqdataset we use. Should probably tune
        prefetch_size: int = 32,
    ):
        """

        TODO: document this

        Args:
            Batch (hax.Axis): The batch axis
            data (AsyncDataset[Ex]): The dataset to load from
            max_buffered_batches (Optional[int]): The maximum number of batches to buffer. If None, the buffer is unbounded.
             If <0, the buffer is disabled and single threaded operation is used.
            axis_resources (Optional[ResourceMapping]): axis mapping
            prefetch_size (int): The number of batches to prefetch at once
            mesh (Mesh): The mesh to use

        """
        self.max_buffered_batches = max_buffered_batches
        self.prefetch_size = prefetch_size
        self.axis_resources = axis_resources
        self.data_store = data
        self.mesh = mesh
        self.Batch = Batch

        with local_cpu_mesh():
            # It's important that all data loading happens CPU side. We might relax this one day.
            self._ex_leaves, self._ex_structure = jax.tree_flatten(
                blocking_wait(self.data_store.getitem_async(0)), is_leaf=is_named_array
            )

        local_device_indices, local_indices = self._compute_local_device_indices()

        self._local_device_indices: dict[jax.Device, range] = local_device_indices
        # this is just the flat indices
        self._local_indices: list[int] = local_indices

    def _compute_local_device_indices(self):
        sharding: jax.sharding.Sharding = hax.partitioning.sharding_for_axis(
            self.Batch.name, self.axis_resources, self.mesh
        )
        # this is a map from devices to the slice of the array that they contain (in the global array)
        local_indices_map = sharding.addressable_devices_indices_map((self.batch_size,))
        # we just want all the indices
        local_device_indices: dict[jax.Device, range] = {
            device1: range(*idx[0].indices(self.batch_size))
            for device1, idx in local_indices_map.items()
            if idx is not None
        }
        local_indices: list[int] = []
        for device, indices in local_device_indices.items():
            local_indices.extend(indices)
        return local_device_indices, local_indices

    @property
    def batch_size(self):
        return self.Batch.size

    def __iter__(self):
        return self.iter_from_step(None)

    def iter_from_step(self, start_from_batch: Optional[int] = None):
        # sometimes we pass in an array for the start_from_batch, so we need to check for that
        start_from_batch = int(start_from_batch) if start_from_batch is not None else None
        return DataLoaderIterator(self, start_from_batch=start_from_batch)


class DataLoaderIterator(Iterator[Ex]):
    def __init__(self, data_loader: DataLoader, start_from_batch: Optional[int] = None):
        self.dl = data_loader
        self._start_from_batch = start_from_batch
        self.mapping = self.dl.axis_resources
        if self.mapping is None:
            self.mapping = hax.partitioning.current_thread_local_mapping()

        buffered_batches = self.dl.max_buffered_batches
        self._batches: Iterator[Ex]
        if buffered_batches == 0:
            self._batches = AsyncIteratorWrapper(self._produce_batches())
        else:
            self._batches = _JaxCpuBackgroundIterator(self._produce_batches, max_capacity=buffered_batches)

    def __next__(self):
        time_start = time.time()
        individual_data_batch = next(self._batches)
        data_for_this_batch = {index: datum for index, datum in zip(self.dl._local_indices, individual_data_batch)}
        batch = self._batchify_local_data(data_for_this_batch)

        time_end = time.time()
        if (time_end - time_start) > 0.5:
            logger.info(f"Prefetch wasn't fast enough: {time_end - time_start:.3f}")
        return batch

    async def _produce_batches(self):
        batch_number = self._start_from_batch or 0
        done = False
        while not done:
            target_next_batch_number = batch_number + self.dl.prefetch_size
            max_achievable_batch_number = await self._dataset_get_available_batch_number(target_next_batch_number)
            if max_achievable_batch_number < target_next_batch_number:
                done = True

            next_batch_numbers = list(range(batch_number, min(target_next_batch_number, max_achievable_batch_number)))

            if len(next_batch_numbers) == 0:
                break

            batch_number = next_batch_numbers[-1] + 1

            async for batch in self._retrieve_batches(next_batch_numbers):
                yield batch

    async def _dataset_get_available_batch_number(self, target_max_batch_number: int) -> int:
        if self.dl.data_store.is_finite():
            next_end = (target_max_batch_number + 1) * self.dl.batch_size
            available_len = await self.dl.data_store.wait_until_len_at_least(next_end)
            max_achievable_batch_number = available_len // self.dl.batch_size

            return max_achievable_batch_number

        return target_max_batch_number

    async def _retrieve_batches(self, batch_numbers: list[int]):
        with local_cpu_mesh():
            time_start = time.time()
            individual_datums_for_each_batch = await self._do_retrieve_batch_of_batches(batch_numbers)
            # reshape to be per batch
            time_end = time.time()
            logger.debug(f"Time to get {len(batch_numbers)} batches: {time_end - time_start:.3f}")

            for data in individual_datums_for_each_batch:
                yield data

    def _batchify_local_data(self, data_for_this_batch: dict[int, Array]):
        cache: dict[tuple[int, int], list[Array | hax.NamedArray]] = {}

        def get_local_batch(begin: int, end: int) -> list:
            if (begin, end) in cache:
                return cache[(begin, end)]

            # TODO: if we ever do "big data" (i.e. huge examples) we might want to be able to load part of an example
            # which will require support from the datastore (i.e. tensorstore)
            device_batch = _stack_tree(self.dl.Batch.name, [data_for_this_batch[i] for i in range(begin, end)])
            batch_leaves = hax.tree_util.tree_leaves(device_batch)

            cache[(begin, end)] = batch_leaves

            return batch_leaves

        def get_local_data_for_leaf(indices: _TensorSliceIndex, leaf_index: int) -> Array:
            batch_slice = indices[0]
            begin, end, stride = batch_slice.indices(self.dl.batch_size)
            if stride != 1:
                raise ValueError("Stride must be 1")

            leaf_data = get_local_batch(begin, end)[leaf_index]

            if isinstance(leaf_data, hax.NamedArray):
                # select out the batch axis
                batch_index = index_where(lambda ax: ax.name == self.dl.Batch.name, leaf_data.axes)
                new_indices = list(indices)
                new_indices[batch_index] = slice(None)
                return leaf_data.array[tuple(new_indices)]
            else:
                other_indices = indices[1:]
                if all(idx == slice(None) for idx in other_indices):
                    return leaf_data
                else:
                    return leaf_data[(..., *other_indices)]

        def make_global_array_for_leaf(leaf_index, item_leaf_shape: ShapeSpec | NamedShapeSpec):
            def get_data(indices):
                return get_local_data_for_leaf(indices, leaf_index)

            raw_array = jax.make_array_from_callback(
                to_raw_shape(item_leaf_shape),
                jax.sharding.NamedSharding(self.dl.mesh, self._pspec_for(item_leaf_shape)),
                get_data,
            )
            if isinstance(item_leaf_shape, NamedShapeSpec):
                return hax.NamedArray(raw_array, item_leaf_shape.shape)
            else:
                return raw_array

        gda_leaves = [
            make_global_array_for_leaf(leaf_index, _batchified_shape(self.dl.Batch, item_leaf))
            for leaf_index, item_leaf in enumerate(self.dl._ex_leaves)
        ]
        gda_tree = jax.tree.unflatten(self.dl._ex_structure, gda_leaves)
        return gda_tree

    async def _do_retrieve_batch_of_batches(self, batch_numbers):
        indices_for_this_batch_of_batches: list[int] = []
        for bn in batch_numbers:
            indices_this_batch = range(bn * self.dl.batch_size, (bn + 1) * self.dl.batch_size, 1)
            indices_this_batch_this_process = [indices_this_batch[i] for i in self.dl._local_indices]
            indices_for_this_batch_of_batches.extend(indices_this_batch_this_process)
        individual_datums = await self.dl.data_store.get_batch(indices_for_this_batch_of_batches)
        individual_datums_for_each_batch = list(batched(individual_datums, len(self.dl._local_indices)))
        return individual_datums_for_each_batch

    def _pspec_for(self, shape_spec: ShapeSpec | NamedShapeSpec) -> PartitionSpec:
        if isinstance(shape_spec, ShapeSpec):  # type: ignore
            batch_name = hax.partitioning.physical_axis_name(self.dl.Batch, self.dl.axis_resources)
            return PartitionSpec(batch_name, *((None,) * (len(shape_spec.shape) - 1)))
        else:
            return hax.partitioning.pspec_for_axis(shape_spec.shape, self.dl.axis_resources)  # type: ignore


def _batchified_shape(Batch, leaf: hax.NamedArray | Array) -> ShapeSpec | NamedShapeSpec:
    if is_named_array(leaf):
        return NamedShapeSpec((Batch,) + leaf.axes, leaf.dtype)
    else:
        return ShapeSpec((Batch.size,) + leaf.shape, leaf.dtype)


class _JaxCpuBackgroundIterator(BackgroundIterator[Ex]):
    """
    We want the thread to only use the CPU device.
    """

    def __init__(self, producer_fn: Callable[[], Iterator[Ex] | AsyncIterator[Ex]], max_capacity: Optional[int]):
        super().__init__(producer_fn, max_capacity)

    def _fill_queue_with_batches(self):
        with local_cpu_mesh():
            super()._fill_queue_with_batches()


@functools.partial(jax.jit, static_argnums=(0,))
def _stack_tree(batch_name, individual_datums):
    def _stack_leaves_unchecked(*leaves):
        if is_named_array(leaves[0]):
            return hax.stack(batch_name, leaves)
        else:
            return jnp.stack(leaves)

    return jax.tree.map(_stack_leaves_unchecked, *individual_datums, is_leaf=is_named_array)


def check_sharded_consistency(tree: PyTree, check_disjoint_indices_are_different: bool = False):
    """Checks the following consistency conditions on an array:
    - all replicas have the same data
    - if check_disjoint_indices_are_different is True, then all shards with disjoint indices have different data
    """
    # index is a tuple[slice, ...], slices are obnoxiously not hashable so we have to convert to tuple

    def check_array(array):
        def _to_tuple(index: Tuple[slice, ...]) -> Tuple[Tuple[int, int], ...]:
            my_indices: Tuple[Tuple[int, int], ...] = tuple(
                s.indices(axis_size)[0:2] for axis_size, s in zip(array.shape, index)
            )

            return my_indices

        replicas_by_index = defaultdict(list)
        for shard in array.global_shards:
            replicas_by_index[_to_tuple(shard.index)].append(shard)

        # global shards is not necessarily sorted consistently, so we have to sort the indices
        sorted_indices = sorted(replicas_by_index.keys())

        # ok now get canonical versions of each index
        replica_0_arrays = {}

        for index in sorted_indices:
            shards = replicas_by_index[index]
            try:
                leader = next(s for s in shards if s.replica_id == 0)
            except StopIteration:
                raise ValueError("No replica 0 for index", index)

            data = leader.data
            if data is None:
                shard_shape = [s[1] - s[0] for s in index]
                data = jnp.zeros(shard_shape, dtype=array.dtype)

            replica_0_array = multihost_utils.broadcast_one_to_all(data, is_source=leader.data is not None)
            replica_0_arrays[index] = replica_0_array

        for shard in array.addressable_shards:
            replica_0_array = replica_0_arrays[_to_tuple(shard.index)]
            assert shard.data is not None

            if not jnp.array_equal(shard.data, replica_0_array, equal_nan=True):
                raise ValueError("Shard data does not match replica 0 data", shard, replica_0_array)

            if check_disjoint_indices_are_different:
                for other_index, other_array in replica_0_arrays.items():
                    if other_index == _to_tuple(shard.index):
                        continue

                    if shard.index != other_index:
                        if jnp.array_equal(shard.data, other_array, equal_nan=True):
                            raise ValueError(
                                "Shard data is the same as another shard with disjoint indices", shard, other_array
                            )

    for leaf in jtu.tree_leaves(tree):
        check_array(leaf)



import os.path
import tempfile

import equinox as eqx
import jax
import numpy as np
import optax
from chex import assert_trees_all_close
from transformers import AutoModelForCausalLM

import haliax as hax
import haliax.nn as hnn
from haliax.quantization import DefaultDotGeneralOp, DotGeneralOp

from levanter.checkpoint import Checkpointer
from levanter.compat.hf_checkpoints import HFCheckpointConverter
from levanter.lora import (
    LoraConfig,
    LoraLinear,
    lora_state_dict,
    lora_trainable_params_filter,
    loraize,
    merge_lora_modules,
    save_merged_hf_model,
    save_peft_pretrained,
)
from levanter.models.attention import AttentionMask
from levanter.models.gpt2 import Gpt2Config, Gpt2LMHeadModel
from levanter.trainer import StepInfo
from levanter.trainer_state import TrainerState
from levanter.utils.tree_utils import inference_mode
from test_utils import skip_if_no_torch


In = hax.Axis("In", 10)
Mid = hax.Axis("Mid", 20)
Out = hax.Axis("Out", 5)


def test_loraize_simple():

    k0 = jax.random.PRNGKey(0)
    k1 = jax.random.PRNGKey(1)

    class Module(eqx.Module):
        first: hnn.Linear
        second: hnn.Linear

        def __call__(self, x):
            return self.second(self.first(x))

    module = Module(first=hnn.Linear.init(In, Mid, key=k0), second=hnn.Linear.init(Mid, Out, key=k1))

    loraized = loraize(module, LoraConfig(r=8, target_modules=["first"]), key=k0)
    assert isinstance(loraized.first, LoraLinear)
    assert isinstance(loraized.second, hnn.Linear)

    loraized = loraize(module, LoraConfig(r=8, target_modules=["second"]), key=k0)
    assert isinstance(loraized.first, hnn.Linear)
    assert isinstance(loraized.second, LoraLinear)

    input = hax.random.normal(k0, (In,))
    assert not hax.all(hax.isclose(module(input), loraized(input)))


def test_lora_scan_layers():
    class Module(eqx.Module):
        first: hnn.Linear
        second: hnn.Linear

        def __call__(self, x):
            return self.second(self.first(x))

        @staticmethod
        def init(*, key):
            k1, k2 = jax.random.split(key)
            first = hnn.Linear.init(In, Mid, key=k1, out_first=True)
            second = hnn.Linear.init(Mid, In, key=k2, out_first=True)
            return Module(first, second)

    Layers = hax.Axis("Layers", 3)

    k0 = jax.random.PRNGKey(0)
    module: hnn.Stacked[Module] = hnn.Stacked.init(Layers, Module)(key=jax.random.split(k0, 3))

    loraized = loraize(module, LoraConfig(r=8, target_modules=["first"]), key=k0)
    assert isinstance(loraized, hnn.Stacked)
    assert isinstance(loraized.stacked.first, LoraLinear)
    assert isinstance(loraized.stacked.second, hnn.Linear)

    assert loraized.stacked.first.lora.lora_A.weight.axes == (Layers, hax.Axis("LORA_R", 8), In)
    assert loraized.stacked.first.lora.lora_B.weight.axes == (Layers, Mid, hax.Axis("LORA_R", 8))

    assert loraized.stacked.second.weight.axes == (Layers, In, Mid)
    input = hax.random.normal(k0, (In,))
    assert not hax.all(hax.isclose(module.fold(input), loraized.fold(input)))


@skip_if_no_torch
def test_lora_peft_integration():
    import peft
    from transformers import AutoModelForCausalLM

    base_hf_model = AutoModelForCausalLM.from_pretrained("stanford-crfm/expanse-gpt2-small-x777")
    peft_config = peft.tuners.LoraConfig(
        base_model_name_or_path="stanford-crfm/expanse-gpt2-small-x777",
        peft_type="lora",
    )
    model = peft.get_peft_model(base_hf_model, peft_config)

    from peft.utils.save_and_load import get_peft_model_state_dict

    hf_dict = get_peft_model_state_dict(model)

    converter = Gpt2Config().hf_checkpoint_converter()

    lev_model = converter.load_pretrained(converter.default_config.model_type, "stanford-crfm/expanse-gpt2-small-x777")

    lora_lev_model = loraize(lev_model, LoraConfig(r=8, target_modules=["c_attn"]), key=jax.random.PRNGKey(0))
    # for some dumb reason, the hf state dict starts with this prefix
    lev_dict = lora_state_dict(lora_lev_model)

    assert lev_dict.keys() == hf_dict.keys()

    for k, v in lev_dict.items():
        assert v.shape == hf_dict[k].shape


def test_merge_lora():
    class Module(eqx.Module):
        first: hnn.Linear
        second: hnn.Linear

        def __call__(self, x):
            return self.second(self.first(x))

        @staticmethod
        def init(*, key):
            k1, k2 = jax.random.split(key)
            first = hnn.Linear.init(In, Mid, key=k1)
            second = hnn.Linear.init(Mid, In, key=k2)
            return Module(first, second)

    Layers = hax.Axis("Layers", 2)

    # tpu matmuls are very imprecise, so we force higher precision
    class PreciseDotGeneralOp(DotGeneralOp):
        def __call__(self, lhs, rhs, dimension_numbers, precision=None, preferred_element_type=None):
            return jax.lax.dot_general(
                lhs,
                rhs,
                dimension_numbers,
                precision=jax.lax.Precision.HIGHEST,
                preferred_element_type=preferred_element_type,
            )

    k0 = jax.random.PRNGKey(0)
    module: hnn.Stacked[Module] = hnn.Stacked.init(Layers, Module)(key=jax.random.split(k0, Layers.size))

    loraized = loraize(module, LoraConfig(r=8, target_modules=["second"]), key=k0)
    assert isinstance(loraized, hnn.Stacked)

    merged = merge_lora_modules(loraized)

    assert isinstance(merged, hnn.Stacked)

    def replace_dot_general(x):
        if isinstance(x, DefaultDotGeneralOp):
            return PreciseDotGeneralOp()
        return x

    merged = jax.tree.map(replace_dot_general, merged, is_leaf=lambda x: isinstance(x, DefaultDotGeneralOp))
    loraized = jax.tree.map(replace_dot_general, loraized, is_leaf=lambda x: isinstance(x, DefaultDotGeneralOp))

    input = hax.random.normal(k0, (In,))
    # light tolerances for TPU
    assert_trees_all_close(merged.fold(input), loraized.fold(input), rtol=1e-3, atol=3e-3)


@skip_if_no_torch
def test_lora_load_in_peft():
    import torch

    converter: HFCheckpointConverter = Gpt2Config().hf_checkpoint_converter()
    config = Gpt2Config(seq_len=128, num_layers=2, num_heads=2)
    Vocab = converter.Vocab

    model = Gpt2LMHeadModel.init(Vocab, config=config, key=jax.random.PRNGKey(0))
    model = inference_mode(model, True)

    input = hax.random.randint(jax.random.PRNGKey(0), config.Pos, 0, Vocab.size)
    torch_input = torch.tensor(np.array(input.array), dtype=torch.long).reshape((1, -1))

    causal_mask = AttentionMask.causal()

    with tempfile.TemporaryDirectory() as tmpdir:
        from peft import PeftConfig, PeftModel

        converter.save_pretrained(model, f"{tmpdir}/model")

        lora_config = LoraConfig(r=8, target_modules=["c_attn"])
        loraized = loraize(model, lora_config, key=jax.random.PRNGKey(0))
        save_peft_pretrained(loraized, lora_config, f"{tmpdir}/model", f"{tmpdir}/loraized")
        peft_config = PeftConfig.from_pretrained(f"{tmpdir}/loraized")

        hf_model = AutoModelForCausalLM.from_pretrained(peft_config.base_model_name_or_path).cpu()
        hf_model.eval()
        hf_out = hf_model(torch_input)
        hf_out = hf_out.logits.detach().numpy()

        lev_out = model(input, attn_mask=causal_mask)
        lev_out = np.array(lev_out.array)
        assert np.allclose(lev_out, hf_out, atol=1e-4)

        # load with peft
        hf_lora_model = PeftModel.from_pretrained(hf_model, f"{tmpdir}/loraized").cpu()

        lev_lora_out = loraized(input, attn_mask=causal_mask)
        lev_lora_out = np.array(lev_lora_out.array)

        hf_lora_model.eval()
        hf_lora_out = hf_lora_model(torch_input)
        hf_lora_out = hf_lora_out.logits.detach().numpy()

        assert np.allclose(lev_lora_out, hf_lora_out, atol=1e-4)
        assert not np.allclose(lev_lora_out, hf_out, atol=1e-4)


@skip_if_no_torch
def test_lora_merged_load_in_hf():
    import torch

    converter: HFCheckpointConverter = Gpt2Config().hf_checkpoint_converter()
    config = Gpt2Config(seq_len=128, num_layers=2, num_heads=2)
    Vocab = converter.Vocab

    model = Gpt2LMHeadModel.init(Vocab, config=config, key=jax.random.PRNGKey(0))
    model = inference_mode(model, True)

    input = hax.random.randint(jax.random.PRNGKey(0), config.Pos, 0, Vocab.size)
    torch_input = torch.tensor(np.array(input.array), dtype=torch.long).reshape((1, -1))

    causal_mask = AttentionMask.causal()

    with (tempfile.TemporaryDirectory() as tmpdir):
        converter.save_pretrained(model, f"{tmpdir}/model")

        lora_config = LoraConfig(r=8, target_modules=["c_attn"])
        loraized = loraize(model, lora_config, key=jax.random.PRNGKey(0))
        save_merged_hf_model(loraized, converter, f"{tmpdir}/loraized")

        hf_model = AutoModelForCausalLM.from_pretrained(f"{tmpdir}/model").cpu()
        hf_model.eval()
        hf_out = hf_model(torch_input)
        hf_out = hf_out.logits.detach().numpy()

        lev_out = model(input, attn_mask=causal_mask)
        lev_out = np.array(lev_out.array)
        assert np.allclose(lev_out, hf_out, atol=1e-4)

        # load merged model with hf
        hf_lora_model = AutoModelForCausalLM.from_pretrained(f"{tmpdir}/loraized").cpu()

        lev_lora_out = loraized(input, attn_mask=causal_mask)
        lev_lora_out = np.array(lev_lora_out.array)

        hf_lora_model.eval()
        hf_lora_out = hf_lora_model(torch_input)
        hf_lora_out = hf_lora_out.logits.detach().numpy()

        assert np.allclose(lev_lora_out, hf_lora_out, atol=1e-4)
        assert not np.allclose(lev_lora_out, hf_out, atol=1e-4)


def test_lora_works_with_checkpointer():
    with tempfile.TemporaryDirectory() as tempdir:
        k0 = jax.random.PRNGKey(0)
        k1 = jax.random.PRNGKey(1)

        class Module(eqx.Module):
            first: hnn.Linear
            second: hnn.Linear

            def __call__(self, x):
                return self.second(self.first(x))

        module = Module(first=hnn.Linear.init(In, Mid, key=k0), second=hnn.Linear.init(Mid, Out, key=k1))

        loraized = loraize(module, LoraConfig(r=8, target_modules=["first"]), key=k0)
        lora_filter = lora_trainable_params_filter(loraized)

        optimizer = optax.adam(1e-3)

        trainer_state = TrainerState.init(optimizer, loraized, key=k0, is_trainable=lora_filter)
        info = StepInfo(trainer_state, 0.0, 0.0)

        checkpointer = Checkpointer(tempdir, None, [])
        checkpointer.save_checkpoint(info, "loraized")

        checkpointer.wait_until_finished()

        # check on disk that we didn't serialize the non-loraized parameters
        if os.path.exists(f"{tempdir}/loraized/model/first/wrapped"):
            assert False

        assert os.path.exists(f"{tempdir}/loraized/model/first/lora/lora_A")



import asyncio
import time

import pytest
import ray

from levanter.utils.actor_pool import AutoScalingActorPool, PoolWorkerBase
from levanter.utils.py_utils import logical_cpu_core_count


@ray.remote
class TestActor(PoolWorkerBase):
    def __init__(self):
        self.node_id = ray.get_runtime_context().get_node_id()

    def get_node_id(self):
        return self.node_id

    def double(self, v):
        return 2 * v


@ray.remote
class BlockerActor(PoolWorkerBase):
    def __init__(self):
        self.node_id = ray.get_runtime_context().get_node_id()
        self.unblocked = False
        self.unblock_event = asyncio.Event()

    def get_node_id(self):
        return self.node_id

    async def block(self):
        if not self.unblocked:
            await self.unblock_event.wait()

    async def unblock(self):
        self.unblocked = True
        self.unblock_event.set()


@ray.remote
class BlockingTestActor(PoolWorkerBase):
    def __init__(self, blocker):
        self.node_id = ray.get_runtime_context().get_node_id()
        self.blocker = blocker

    def get_node_id(self):
        return self.node_id

    def double(self, v, bypass_blocker=False):
        if not bypass_blocker:
            ray.get(self.blocker.block.remote())
        return 2 * v


# Helper function to create a TestActor
def create_test_actor():
    return TestActor.remote()


def create_test_actor_blocker(blocker_handle):
    return BlockingTestActor.remote(blocker_handle)


def setup_module(module):
    ray.init(
        "local", num_cpus=max(2 * logical_cpu_core_count(), 8), ignore_reinit_error=True
    )  # 2x cpu count is faster on my m1


def teardown_module(module):
    ray.shutdown()


@pytest.mark.asyncio
async def test_basic_submit():
    pool = AutoScalingActorPool(create_test_actor, min_size=1, max_size=4)
    results = [pool.submit(lambda a, v: a.double.remote(v), i) for i in range(4)]
    results = [await r for r in results]

    assert results == [0, 2, 4, 6]


@pytest.mark.asyncio
async def test_basic_submit_no_idle():
    pool = AutoScalingActorPool(create_test_actor, min_size=0, max_size=4)
    results = [pool.submit(lambda a, v: a.double.remote(v), i) for i in range(4)]
    results = [await r for r in results]

    assert results == [0, 2, 4, 6]


@pytest.mark.asyncio
async def test_basic_functionality():
    pool = AutoScalingActorPool(create_test_actor, min_size=1, max_size=4)
    results = list(await pool.map(lambda a, v: a.double.remote(v), [1, 2, 3, 4]))
    assert results == [2, 4, 6, 8]


@pytest.mark.asyncio
async def test_scaling_up():
    blocker = BlockerActor.remote()
    pool = AutoScalingActorPool(lambda: create_test_actor_blocker(blocker), min_size=1, max_size=4)
    f1 = pool.submit(lambda a, v: a.double.remote(v), 1)
    f2 = pool.submit(lambda a, v: a.double.remote(v), 2)
    f3 = pool.submit(lambda a, v: a.double.remote(v, True), 3)
    f4 = pool.submit(lambda a, v: a.double.remote(v, True), 4)

    shield_f2 = asyncio.shield(f2)
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(shield_f2, timeout=0.1)

    assert (await asyncio.gather(f3, f4)) == [6, 8]

    await blocker.unblock.remote()
    # assert (await asyncio.gather(f1, f2)) == [2, 4]
    assert (await f1) == 2
    assert (await f2) == 4


@pytest.mark.asyncio
async def test_scaling_down():
    pool = AutoScalingActorPool(create_test_actor, min_size=1, max_size=4)
    await pool.submit(lambda a, v: a.double.remote(v), 1)
    await pool.submit(lambda a, v: a.double.remote(v), 2)
    await pool.submit(lambda a, v: a.double.remote(v), 3)
    await pool.submit(lambda a, v: a.double.remote(v), 4)
    results = await asyncio.gather(
        pool.submit(lambda a, v: a.double.remote(v), 1),
        pool.submit(lambda a, v: a.double.remote(v), 2),
        pool.submit(lambda a, v: a.double.remote(v), 3),
        pool.submit(lambda a, v: a.double.remote(v), 4),
    )
    assert results == [2, 4, 6, 8]
    assert len(pool._idle_actors) == 1
    assert len(pool._busy_actors) == 0


@pytest.mark.asyncio
async def test_push_pop_idle():
    pool = AutoScalingActorPool(create_test_actor, min_size=1, max_size=4)
    await pool.submit(lambda a, v: a.double.remote(v), 1)
    actor = pool.pop_idle()
    assert actor is not None
    pool.push(actor)
    assert len(pool._idle_actors) == 1


@pytest.mark.asyncio
async def test_submit_with_no_idle_actors():
    blocker = BlockerActor.remote()
    pool = AutoScalingActorPool(lambda: create_test_actor_blocker(blocker), min_size=1, max_size=4)
    futs = [pool.submit(lambda a, v: a.double.remote(v), i) for i in range(4)]
    f5 = pool.submit(lambda a, v: a.double.remote(v), 5)
    await _sleep_until(lambda: pool.num_pending_tasks == 1, timeout=10)
    await blocker.unblock.remote()
    await asyncio.gather(*futs)
    assert (await f5) == 10


async def _sleep_until(condition, timeout=5, message="Condition not met within timeout"):
    start = time.time()
    while not condition():
        if time.time() - start > timeout:
            pytest.fail(message)
        await asyncio.sleep(0.1)



import dataclasses
from dataclasses import dataclass
from functools import partial
from typing import Callable, Dict, Optional, Type

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

import haliax as hax
import haliax.jax_utils
import haliax.nn as hnn
from haliax import Axis, NamedArray
from haliax.jax_utils import named_call, shaped_rng_split
from haliax.nn.scan import Stacked
from haliax.state_dict import ModuleWithStateDictSerialization

from levanter.compat.hf_checkpoints import HFCheckpointConverter, HFCompatConfig, ModelWithHfSerializationMixin
from levanter.logging import silence_transformer_nag
from levanter.models.asr_model import ASRConfig, ASRMixin
from levanter.models.attention import AttentionBackend, AttentionMask, dot_product_attention
from levanter.models.lm_model import LmConfig


silence_transformer_nag()
from transformers import PretrainedConfig as HfConfig  # noqa: E402
from transformers import WhisperConfig as HfWhisperConfig  # noqa: E402


@LmConfig.register_subclass("whisper")
@dataclass(frozen=True)
class WhisperConfig(HFCompatConfig, ASRConfig):
    vocab_size: int = 51865
    num_mel_bins: int = 80
    encoder_layers: int = 4
    encoder_attention_heads: int = 6
    encoder_ffn_dim: int = 1536
    decoder_layers: int = 4
    decoder_attention_heads: int = 6
    decoder_ffn_dim: int = 1536
    d_model: int = 384

    max_source_positions: int = 1500
    max_length: int = 448

    activation_function: str = "gelu"
    layer_norm_epsilon: float = 1e-5
    use_bias: bool = True

    initializer_range: float = 0.02
    gradient_checkpointing: bool = True

    # Attention-related config
    upcast_attn: bool = True
    use_flash_attention: bool = False
    attn_backend: Optional[AttentionBackend] = None
    flash_attention_block_size: Optional[int] = None

    @property
    def model_type(self) -> Type["WhisperModel"]:
        return WhisperModel

    @property
    def asr_model_type(self) -> Type["WhisperASRModel"]:
        return WhisperASRModel

    def hf_checkpoint_converter(self) -> HFCheckpointConverter["WhisperModel"]:  # type: ignore
        return HFCheckpointConverter(self, "openai/whisper-base", ignore_prefix="model")

    # Axis
    MelPos = property(lambda self: Axis(name="position", size=self.max_source_positions * 2))
    Pos = property(lambda self: Axis(name="position", size=self.max_length))
    KeyPos = property(lambda self: self.Pos.alias("key_position"))
    SourcePos = property(lambda self: Axis(name="position", size=self.max_source_positions))
    Vocab = property(lambda self: Axis(name="vocab_size", size=self.vocab_size))
    Embed = property(lambda self: Axis(name="embed_dim", size=self.d_model))
    EncoderMlp = property(lambda self: Axis(name="mlp_dim", size=self.encoder_ffn_dim))
    EncoderHeads = property(lambda self: Axis(name="heads", size=self.encoder_attention_heads))
    EncoderHeadSize = property(lambda self: Axis(name="head_size", size=self.d_model // self.encoder_attention_heads))
    EncoderLayer = property(lambda self: Axis(name="encoder_layers", size=self.encoder_layers))
    DecoderMlp = property(lambda self: Axis(name="mlp_dim", size=self.decoder_ffn_dim))
    DecoderHeads = property(lambda self: Axis(name="heads", size=self.decoder_attention_heads))
    DecoderHeadSize = property(lambda self: Axis(name="head_size", size=self.d_model // self.decoder_attention_heads))
    DecoderLayer = property(lambda self: Axis(name="decoder_layers", size=self.decoder_layers))
    Mels = property(lambda self: Axis(name="n_mels", size=self.num_mel_bins))

    def to_hf_config(self, vocab_size, config_overrides=None):
        if config_overrides is None:
            config_overrides = {}

        return HfWhisperConfig(
            vocab_size=self.vocab_size,
            num_mel_bins=self.num_mel_bins,
            encoder_layers=self.encoder_layers,
            encoder_attention_heads=self.encoder_attention_heads,
            decoder_layers=self.decoder_layers,
            decoder_attention_heads=self.decoder_attention_heads,
            decoder_ffn_dim=self.decoder_ffn_dim,
            encoder_ffn_dim=self.encoder_ffn_dim,
            activation_function=self.activation_function,
            max_source_positions=self.max_source_positions,
            d_model=self.d_model,
        )

    @classmethod
    def from_hf_config(cls, hf_config: HfConfig):
        return cls(
            vocab_size=hf_config.vocab_size,
            num_mel_bins=hf_config.num_mel_bins,
            encoder_layers=hf_config.encoder_layers,
            encoder_attention_heads=hf_config.encoder_attention_heads,
            decoder_layers=hf_config.decoder_layers,
            decoder_attention_heads=hf_config.decoder_attention_heads,
            decoder_ffn_dim=hf_config.decoder_ffn_dim,
            encoder_ffn_dim=hf_config.encoder_ffn_dim,
            activation_function=hf_config.activation_function,
            max_source_positions=hf_config.max_source_positions,
            d_model=hf_config.d_model,
        )


class WhisperMlp(eqx.Module):
    fc1: hnn.Linear  # projection from Embed to Intermediate (typically 4x Embed)
    fc2: hnn.Linear  # projection from Intermediate to Embed
    act: Callable = eqx.static_field()

    @staticmethod
    def init(Embed: Axis, Mlp: Axis, activation_fn, *, key, use_bias: bool = True) -> "WhisperMlp":
        k_fc, k_proj = haliax.jax_utils.maybe_rng_split(key, 2)
        fc1 = hnn.Linear.init(Out=Mlp, In=Embed, key=k_fc, use_bias=use_bias, out_first=False)
        fc2 = hnn.Linear.init(Out=Embed, In=Mlp, key=k_proj, use_bias=use_bias, out_first=False)
        if isinstance(activation_fn, str):
            activation_fn = ACT2FN[activation_fn]
        act = activation_fn  # type: ignore

        return WhisperMlp(fc1, fc2, act)

    @named_call
    def __call__(self, x: NamedArray, *, key=None):
        k1, k2 = haliax.jax_utils.maybe_rng_split(key, 2)
        x = self.fc1(x, key=k1)
        x = self.act(x)
        x = self.fc2(x, key=k2)
        return x


class WhisperAttention(eqx.Module):
    config: WhisperConfig = eqx.static_field()

    q_proj: hnn.Linear  # input projection from [embed] -> [q, heads, head_dim]
    k_proj: hnn.Linear  # input projection from [embed] -> [k, heads, head_dim]
    v_proj: hnn.Linear  # input projection from [embed] -> [v, heads, head_dim]

    out_proj: hnn.Linear  # output projection from [heads, head_dim] -> [embed]
    inference: bool

    @staticmethod
    def init(Heads: Axis, HeadSize: Axis, config: WhisperConfig, *, key) -> "WhisperAttention":
        use_bias = config.use_bias
        Embed = config.Embed

        k_q, k_k, k_v, k_out = haliax.jax_utils.maybe_rng_split(key, 4)
        q_proj = hnn.Linear.init(In=Embed, Out=(Heads, HeadSize), key=k_q, use_bias=use_bias, out_first=False)
        k_proj = hnn.Linear.init(In=Embed, Out=(Heads, HeadSize), key=k_k, use_bias=False, out_first=False)
        v_proj = hnn.Linear.init(In=Embed, Out=(Heads, HeadSize), key=k_v, use_bias=use_bias, out_first=False)
        out_proj = hnn.Linear.init(In=(Heads, HeadSize), Out=Embed, key=k_out, use_bias=use_bias, out_first=False)

        return WhisperAttention(config, q_proj, k_proj, v_proj, out_proj, inference=False)

    @named_call
    def __call__(self, x: NamedArray, xa: Optional[NamedArray] = None, mask: Optional[AttentionMask] = None, *, key):
        k_k, k_v, k_q, k_out, k_drop = hax.jax_utils.maybe_rng_split(key, 5)
        q = self.q_proj(x, key=k_q).rearrange((..., "heads", "position", "head_size"))
        kv_in = x if xa is None else xa
        k = self.k_proj(kv_in, key=k_k).rearrange((..., "heads", "position", "head_size"))
        v = self.v_proj(kv_in, key=k_v).rearrange((..., "heads", "position", "head_size"))

        # Rename k and v's Pos as haliax doesn't support unnamed axes or duplicate axes
        k = k.rename({"position": "key_position"})
        v = v.rename({"position": "key_position"})

        attn_output = dot_product_attention(
            "position",
            "key_position",
            "head_size",
            q,
            k,
            v,
            mask=mask,
            inference=self.inference,
            use_flash=self.config.use_flash_attention,
            flash_block_size=self.config.flash_attention_block_size,
            prng=k_drop,
            attention_dtype=jnp.float32 if self.config.upcast_attn else None,
        )

        attn_output = attn_output.astype(x.dtype)
        attn_output = self.out_proj(attn_output, key=k_out)

        return attn_output


class WhisperLayer(ModuleWithStateDictSerialization, eqx.Module):
    self_attn: WhisperAttention
    attn_ln: hnn.LayerNorm

    encoder_attn: Optional[WhisperAttention]
    encoder_attn_ln: Optional[hnn.LayerNorm]

    mlp: WhisperMlp
    mlp_ln: hnn.LayerNorm

    @staticmethod
    def init(
        Heads: Axis, HeadSize: Axis, Mlp: Axis, config: WhisperConfig, has_cross: bool = True, *, key
    ) -> "WhisperLayer":
        k_attn, k_cross, k_mlp = haliax.jax_utils.maybe_rng_split(key, 3)

        attn_ln = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)
        self_attn = WhisperAttention.init(Heads, HeadSize, config, key=k_attn)

        if has_cross:
            encoder_attn_ln = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)
            encoder_attn = WhisperAttention.init(Heads, HeadSize, config, key=k_attn)
        else:
            encoder_attn_ln = None
            encoder_attn = None

        mlp = WhisperMlp.init(config.Embed, Mlp, config.activation_function, key=k_mlp, use_bias=config.use_bias)
        mlp_ln = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)

        return WhisperLayer(self_attn, attn_ln, encoder_attn, encoder_attn_ln, mlp, mlp_ln)

    @named_call
    def __call__(self, x: NamedArray, xa: Optional[NamedArray] = None, mask: Optional[AttentionMask] = None, *, key):
        k1, k2, k3 = haliax.jax_utils.maybe_rng_split(key, 3)

        attn_output = self.self_attn(self.attn_ln(x), mask=mask, key=k1)
        x = x + attn_output

        if self.encoder_attn:
            xs = x
            if self.encoder_attn_ln:
                xs = self.encoder_attn_ln(xs)
            x = x + self.encoder_attn(xs, xa, key=k2)

        ff_output = self.mlp(self.mlp_ln(x), key=k3)
        x = x + ff_output

        return x

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {
            "mlp": None,
            "mlp_ln": "final_layer_norm",
            "attn_ln": "self_attn_layer_norm",
            "encoder_attn_ln": "encoder_attn_layer_norm",
        }


class WhisperTransformer(ModuleWithStateDictSerialization):
    layers: Stacked[WhisperLayer]
    Layer: Axis
    layer_norm: hnn.LayerNorm

    @staticmethod
    def init(Layer: Axis, Heads: Axis, HeadSize: Axis, Mlp: Axis, config: WhisperConfig, has_cross: bool, *, key):
        # vectorize the blocks
        layers = Stacked.init(Layer, WhisperLayer, gradient_checkpointing=config.gradient_checkpointing)(
            Heads,
            HeadSize,
            Mlp,
            config,
            has_cross=has_cross,
            key=shaped_rng_split(key, Layer.size),
        )
        layer_norm = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)

        return WhisperTransformer(layers, Layer, layer_norm)

    @named_call
    def __call__(
        self,
        x: NamedArray,
        xa: Optional[NamedArray] = None,
        attn_mask: Optional[AttentionMask] = None,
        *,
        key=None,
    ) -> NamedArray:
        keys = hax.jax_utils.maybe_rng_split(key, self.Layer.size) if key is not None else None
        x = self.layers.fold(x, xa, attn_mask, key=keys)
        x = self.layer_norm(x)

        return x


class WhisperEncoder(ModuleWithStateDictSerialization):
    config: WhisperConfig = eqx.static_field()
    conv1: hnn.Conv
    conv2: hnn.Conv
    act: Callable = eqx.static_field()

    transformer: WhisperTransformer

    @classmethod
    def init(cls, config: WhisperConfig, *, key) -> "WhisperEncoder":
        k_conv1, k_conv2, k_t = haliax.jax_utils.maybe_rng_split(key, 3)

        Len = hax.Axis("position", size=config.SourcePos.size * 2)
        Mid = hax.Axis("mid", config.Embed.size)
        conv1 = hnn.Conv.init(Len, config.Mels, Mid, kernel_size=3, padding=1, key=k_conv1)
        conv2 = hnn.Conv.init(Len, Mid, config.Embed, kernel_size=3, stride=2, padding=1, key=k_conv2)
        if isinstance(config.activation_function, str):
            act = ACT2FN[config.activation_function]  # type: ignore
        else:
            act = config.activation_function

        transformer = WhisperTransformer.init(
            config.EncoderLayer,
            config.EncoderHeads,
            config.EncoderHeadSize,
            config.EncoderMlp,
            config,
            has_cross=False,
            key=k_t,
        )

        return WhisperEncoder(config, conv1, conv2, act, transformer)

    def __call__(self, spec: NamedArray, *, key=None) -> NamedArray:
        k_conv1, k_conv2, k_transformer = haliax.jax_utils.maybe_rng_split(key, 3)
        spec = spec.astype(self.conv1.weight.dtype)
        x = self.act(self.conv1(spec, key=k_conv1))
        x = self.act(self.conv2(x, key=k_conv2))

        seq_len = x.axis_size("position")
        pos_emb = whisper_sinusoids(self.config.Embed, self.config.SourcePos)[self.config.SourcePos, :seq_len]
        x = x + pos_emb

        x = self.transformer(x, key=k_transformer)
        return x

    def resize_vocab(self, new_size: int, key: Optional[PRNGKeyArray] = None) -> "WhisperDecoder":
        new_embeddings = self.embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, embeddings=new_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"transformer": None}


class WhisperDecoderEmbeddings(eqx.Module):
    Vocab: Axis = eqx.static_field()
    config: WhisperConfig = eqx.static_field()

    token_embeddings: hnn.Embedding
    position_embeddings: hnn.Embedding

    @staticmethod
    def init(Vocab: Axis, config: WhisperConfig, *, key) -> "WhisperDecoderEmbeddings":
        k_wte, k_wpe, k_out = haliax.jax_utils.maybe_rng_split(key, 3)

        token_embeddings = hnn.Embedding.init(
            Vocab, config.Embed, key=k_wte, initializer_range=config.initializer_range
        )

        # Whisper Initializes the Positional Embeddings as Empty
        position_embeddings = hnn.Embedding.init(config.Pos, config.Embed, key=k_wpe, initializer_range=0)

        return WhisperDecoderEmbeddings(Vocab, config, token_embeddings, position_embeddings)

    @named_call
    def embed(self, input_ids, *, key):
        input_embeds = self.token_embeddings(input_ids)
        input_Pos = input_ids.resolve_axis("position")
        position_embeds = self.position_embeddings.embed(hax.arange(input_Pos))
        x = input_embeds + position_embeds

        return x

    def unembed(self, x: NamedArray):
        return hax.dot(x, self.token_embeddings.weight, axis="embed_dim")

    def resize_embeddings(self, new_size: int, key: Optional[PRNGKeyArray] = None):
        new_token_embeddings = self.token_embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, Vocab=self.Vocab.resize(new_size), token_embeddings=new_token_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"token_embeddings": "embed_tokens", "position_embeddings": "embed_positions"}


class WhisperDecoder(ModuleWithStateDictSerialization):
    transformer: WhisperTransformer
    embeddings: WhisperDecoderEmbeddings

    @property
    def config(self):
        return self.embeddings.config

    @property
    def Vocab(self) -> Axis:
        return self.embeddings.Vocab

    @property
    def Pos(self) -> Axis:
        return self.config.Pos

    @classmethod
    def init(cls, config: WhisperConfig, *, key) -> "WhisperDecoder":
        k_t, k_embeddings = haliax.jax_utils.maybe_rng_split(key, 2)
        transformer = WhisperTransformer.init(
            config.DecoderLayer,
            config.DecoderHeads,
            config.DecoderHeadSize,
            config.DecoderMlp,
            config,
            has_cross=True,
            key=k_t,
        )
        embeddings = WhisperDecoderEmbeddings.init(config.Vocab, config, key=k_embeddings)

        return WhisperDecoder(transformer, embeddings)

    def __call__(
        self,
        input_ids: NamedArray,
        audio_embeds: NamedArray,
        attn_mask: Optional[AttentionMask] = None,
        *,
        key=None,
    ) -> NamedArray:
        causal_mask = AttentionMask.causal()
        if attn_mask is not None:
            causal_mask = causal_mask & attn_mask
        k_embed, k_transformer = haliax.jax_utils.maybe_rng_split(key, 2)
        x = self.embeddings.embed(input_ids, key=k_embed)
        x = self.transformer(x, audio_embeds, causal_mask, key=k_transformer)
        lm_logits = self.embeddings.unembed(x)

        return lm_logits

    def resize_vocab(self, new_size: int, key: Optional[PRNGKeyArray] = None) -> "WhisperDecoder":
        new_embeddings = self.embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, embeddings=new_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"transformer": None, "embeddings": None}


class WhisperModel(eqx.Module, ModelWithHfSerializationMixin[WhisperConfig]):
    encoder: WhisperEncoder
    decoder: WhisperDecoder

    @property
    def config(self):
        return self.encoder.config

    @property
    def Pos(self) -> Axis:
        return self.config.Pos

    @property
    def Vocab(self) -> Axis:
        return self.decoder.embeddings.Vocab

    def resize_vocab(self, new_size: int, key: Optional[PRNGKeyArray] = None) -> "WhisperModel":
        new_decoder = self.decoder.resize_vocab(new_size, key)
        return dataclasses.replace(self, decoder=new_decoder)

    @classmethod
    def init(cls, Vocab: Axis, config: WhisperConfig, *, key) -> "WhisperModel":
        k_t, k_embeddings = haliax.jax_utils.maybe_rng_split(key, 2)
        encoder = WhisperEncoder.init(config, key=k_embeddings)
        decoder = WhisperDecoder.init(config, key=k_t)

        return cls(encoder, decoder)

    def __call__(
        self,
        mel: NamedArray,
        input_ids: NamedArray,
        attn_mask: Optional[AttentionMask | NamedArray] = None,
        *,
        key=None,
    ) -> NamedArray:
        if attn_mask is not None and not isinstance(attn_mask, AttentionMask):
            attn_mask = AttentionMask.explicit(attn_mask)
        k_encoder, k_decoder = haliax.jax_utils.maybe_rng_split(key, 2)
        audio_features = self.encoder(mel, key=k_encoder)
        lm_logits = self.decoder(input_ids, audio_features, attn_mask=attn_mask, key=k_decoder)

        return lm_logits


class WhisperASRModel(WhisperModel, ASRMixin):
    pass


def whisper_sinusoids(Channels: Axis, SourcePos: Axis, base: int = 10000) -> NamedArray:
    with jax.ensure_compile_time_eval():
        log_timescale_increment = (hax.log(base) / (Channels.size // 2 - 1)).item()
        ChannelHalfSize = Channels.resize(Channels.size // 2)
        inv_timescales = hax.exp(-log_timescale_increment * hax.arange(ChannelHalfSize))

        position_ids: NamedArray = hax.arange(SourcePos)

        freqs = position_ids * inv_timescales.broadcast_axis(SourcePos)
        emb = hax.concatenate(Channels, (hax.sin(freqs), hax.cos(freqs)))
        return emb


ACT2FN: Dict[str, Callable] = {
    "relu": hnn.relu,
    "silu": hnn.silu,
    "swish": hnn.swish,
    "gelu": partial(hnn.gelu, approximate=False),
    "gelu_new": partial(hnn.gelu, approximate=True),
    "quick_gelu": hnn.quick_gelu,
}



import pathlib

import pytest
from git import InvalidGitRepositoryError, NoSuchPathError, Repo

from levanter.tracker.helpers import infer_experiment_git_root


def test_infer_experiment_git_root():
    # make sure this test is running in a git repo
    try:
        repo = Repo(pathlib.Path(__file__), search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        pytest.skip("test not running in a git repo")

    root = infer_experiment_git_root()

    # ensure that 1) this is a git root and 2) this source file is underneath
    assert root is not None
    assert pathlib.Path(root).exists()
    repo = Repo(root)
    assert repo.working_dir == root
    assert pathlib.Path(__file__).is_relative_to(root), f"{__file__} is not relative to {root}"



import dataclasses
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Type, Union

import equinox as eqx
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import PRNGKeyArray

import haliax as hax
import haliax.jax_utils
import haliax.nn as hnn
from haliax import Axis, AxisSpec, NamedArray
from haliax.jax_utils import named_call
from haliax.state_dict import ModuleWithStateDictSerialization, StateDict, with_prefix

from levanter.compat.hf_checkpoints import HFCheckpointConverter, LmWithHfSerializationMixin
from levanter.logging import silence_transformer_nag
from levanter.models.attention import AttentionMask, materialize_mask
from levanter.models.gpt2 import ACT2FN, Gpt2Config, Gpt2Transformer
from levanter.models.lm_model import LmConfig


silence_transformer_nag()
from transformers import PretrainedConfig  # noqa: E402


@LmConfig.register_subclass("backpack")
@dataclass(frozen=True)
class BackpackConfig(Gpt2Config):
    # Backpack-specific terms
    num_senses: int = 16
    sense_intermediate_scale: int = 4

    @property
    def model_type(self) -> Type["BackpackLMHeadModel"]:
        return BackpackLMHeadModel

    def hf_checkpoint_converter(self) -> HFCheckpointConverter["BackpackConfig"]:  # type: ignore
        # We trust this code because it's in our hub repo
        return HFCheckpointConverter(
            self,
            reference_checkpoint="stanford-crfm/levanter-backpack-1b@9face7bd6182155fe3f1a6a5a14ca1c4810bb079",
            trust_remote_code=True,
        )

    # Axes
    SenseHeadDim = property(lambda self: Axis(name="head_dim", size=self.hidden_dim // self.num_senses))
    Senses = property(lambda self: Axis(name="senses", size=self.num_senses))
    SenseIntermediate = property(
        lambda self: Axis(name="concat_senses", size=self.sense_intermediate_scale * self.hidden_dim)
    )

    def to_hf_config(self, vocab_size, config_overrides=None):
        if config_overrides is None:
            config_overrides = {}

        return PretrainedConfig(
            vocab_size=vocab_size,
            n_positions=self.seq_len,
            n_layer=self.num_layers,
            n_head=self.num_heads,
            n_embd=self.hidden_dim,
            initializer_range=self.initializer_range,
            attn_pdrop=self.attn_pdrop,
            embd_pdrop=self.embed_pdrop,
            resid_pdrop=self.resid_pdrop,
            layer_norm_epsilon=self.layer_norm_epsilon,
            activation_function=self.activation_function,
            scale_attn_by_inverse_layer_idx=self.scale_attn_by_inverse_layer_idx,
            reorder_and_upcast_attn=self.upcast_attn,
            num_senses=self.num_senses,
            sense_intermediate_scale=self.sense_intermediate_scale,
            **config_overrides,
        )

    @classmethod
    def from_hf_config(cls, hf_config: PretrainedConfig):
        return cls(
            seq_len=hf_config.n_positions,
            num_layers=hf_config.n_layer,
            num_heads=hf_config.n_head,
            hidden_dim=hf_config.n_embd,
            initializer_range=hf_config.initializer_range,
            attn_pdrop=hf_config.attn_pdrop,
            embed_pdrop=hf_config.embd_pdrop,
            resid_pdrop=hf_config.resid_pdrop,
            layer_norm_epsilon=hf_config.layer_norm_epsilon,
            activation_function=hf_config.activation_function,
            scale_attn_by_inverse_layer_idx=hf_config.scale_attn_by_inverse_layer_idx,
            upcast_attn=hf_config.reorder_and_upcast_attn,
            num_senses=hf_config.num_senses,
            sense_intermediate_scale=hf_config.sense_intermediate_scale,
        )


class BackpackMlp(eqx.Module):
    c_fc: hnn.Linear  # projection from Embed to Intermediate (typically 4x Embed)
    c_proj: hnn.Linear  # projection from Intermediate to Embed
    act: Callable = eqx.static_field()

    @staticmethod
    def init(
        Embed: Axis,
        Mlp: Axis,
        Out: AxisSpec,
        activation_fn: Union[str, Callable],
        *,
        key,
        use_bias: bool = True,
    ) -> "BackpackMlp":
        k_fc, k_proj = jrandom.split(key, 2)
        c_fc = hnn.Linear.init(Out=Mlp, In=Embed, key=k_fc, use_bias=use_bias, out_first=False)
        c_proj = hnn.Linear.init(Out=Out, In=Mlp, key=k_proj, use_bias=use_bias, out_first=False)
        if isinstance(activation_fn, str):
            activation_fn = ACT2FN[activation_fn]
        act = activation_fn  # type: ignore

        return BackpackMlp(c_fc=c_fc, c_proj=c_proj, act=act)

    @named_call
    def __call__(self, x: NamedArray):
        x = self.c_fc(x)
        x = self.act(x)
        x = self.c_proj(x)
        return x


class WeightsOnlyAttention(ModuleWithStateDictSerialization):
    """
    Changes from Gpt2Attention:
    1. No projection; it returns the attention weights
    2. Use SenseHeadDim instead of HeadDim, use Senses instead of Heads
    """

    # No projection
    config: Gpt2Config = eqx.static_field()

    c_attn: hnn.Linear  # input projection from [embed] -> [(q, k, v), heads, head_dim]
    dropout: hnn.Dropout

    @staticmethod
    def init(config: Gpt2Config, *, key) -> "WeightsOnlyAttention":
        Qk = Axis("qk", size=2)
        use_bias = config.use_bias
        Embed = config.Embed

        k_c, _ = jrandom.split(key, 2)
        # NB: out_first=True b/c the torch implementation uses Linear
        c_attn = hnn.Linear.init(
            In=Embed, Out=(Qk, config.Senses, config.SenseHeadDim), key=k_c, use_bias=use_bias, out_first=True
        )
        dropout = hnn.Dropout(config.attn_pdrop)

        return WeightsOnlyAttention(config, c_attn, dropout)

    @named_call
    def __call__(self, x: NamedArray, mask: Optional[AttentionMask | NamedArray], layer_idx, *, key):
        qk_out = self.c_attn(x)
        q, k = qk_out.unbind("qk")

        # Rename k's Pos as haliax doesn't support unnamed axes or duplicate axes
        k = k.rename({"position": "key_position"})
        QPos = q.resolve_axis("position")
        KPos = k.resolve_axis("key_position")

        mask = materialize_mask(mask, q.resolve_axis("position"), k.resolve_axis("key_position"))

        attn_weights = hnn.attention.dot_product_attention_weights(
            "head_dim",
            "key_position",
            q,
            k,
            mask=materialize_mask(mask, QPos, KPos),
            attention_dtype=jnp.float32 if self.config.upcast_attn else None,
        )

        attn_weights = self.dropout(attn_weights, key=key)
        return attn_weights


class NoMixBlock(eqx.Module):
    ln_1: hnn.LayerNorm
    ln_2: hnn.LayerNorm
    mlp: BackpackMlp
    resid_dropout1: hnn.Dropout
    resid_dropout2: hnn.Dropout

    @staticmethod
    def init(config: BackpackConfig, *, key) -> "NoMixBlock":
        k_mlp = jrandom.split(key, 1)[0]

        ln_1 = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon)
        resid_dropout1 = hnn.Dropout(pdrop=config.resid_pdrop)
        resid_dropout2 = hnn.Dropout(pdrop=config.resid_pdrop)
        ln_2 = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon)

        mlp = BackpackMlp.init(
            Embed=config.Embed,
            Mlp=config.Mlp,
            Out=config.Embed,
            activation_fn=config.activation_function,
            key=k_mlp,
            use_bias=config.use_bias,
        )

        return NoMixBlock(ln_1=ln_1, ln_2=ln_2, mlp=mlp, resid_dropout1=resid_dropout1, resid_dropout2=resid_dropout2)

    @named_call
    def __call__(self, hidden_states: NamedArray, residual: NamedArray, *, key):
        k1, k2 = haliax.jax_utils.maybe_rng_split(key, 2)

        residual = self.resid_dropout1(hidden_states, key=k1) + residual
        hidden_states = self.ln_1(residual)
        mlp_out = self.mlp(hidden_states)
        residual = self.resid_dropout2(mlp_out, key=k2) + residual
        hidden_states = self.ln_2(residual)

        return hidden_states


class BackpackSenses(eqx.Module):
    dropout: hnn.Dropout
    block: NoMixBlock
    ln: hnn.LayerNorm
    final_mlp: BackpackMlp

    Pos: Axis = eqx.static_field()

    @staticmethod
    def init(
        config,
        dropout_prob: float,
        *,
        key,
    ):
        k_block, k_mlp = jrandom.split(key, 2)

        dropout = hnn.Dropout(pdrop=dropout_prob)
        block = NoMixBlock.init(config, key=k_block)
        ln = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon)
        final_mlp = BackpackMlp.init(
            Embed=config.Embed,
            Mlp=config.SenseIntermediate,
            Out=(config.Senses, config.Embed),
            activation_fn=config.activation_function,
            key=k_mlp,
            use_bias=config.use_bias,
        )

        return BackpackSenses(
            dropout=dropout,
            block=block,
            ln=ln,
            final_mlp=final_mlp,
            Pos=config.Pos,
        )

    @named_call
    def sense_embed(self, input_embeds, *, key):
        hidden_states = self.ln(input_embeds)
        hidden_states = self.block(hidden_states, input_embeds, key=key)
        senses = self.final_mlp(hidden_states)

        return senses


class BackpackGpt2Embeddings(eqx.Module):
    Vocab: Axis = eqx.static_field()
    config: Gpt2Config = eqx.static_field()

    token_embeddings: NamedArray
    position_embeddings: NamedArray
    dropout: hnn.Dropout

    @staticmethod
    def init(Vocab: Axis, config: Gpt2Config, *, key) -> "BackpackGpt2Embeddings":
        k_wte, k_wpe, k_out = jrandom.split(key, 3)

        token_embeddings = hax.random.normal(k_wte, (Vocab, config.Embed)) * config.initializer_range
        position_embeddings = hax.random.normal(k_wpe, (config.Pos, config.Embed)) * (config.initializer_range / 2)
        dropout = hnn.Dropout(pdrop=config.embed_pdrop)

        return BackpackGpt2Embeddings(Vocab, config, token_embeddings, position_embeddings, dropout)

    @named_call
    def embed_input_ids(self, input_ids: NamedArray) -> NamedArray:
        return self.token_embeddings.take("vocab", input_ids)

    @named_call
    def embed(self, input_ids, *, key):
        input_embeds = self.token_embeddings.take("vocab", input_ids)
        position_embeds = self.position_embeddings
        input_len = input_ids.resolve_axis("position").size
        x = input_embeds + position_embeds["position", hax.dslice(0, input_len)]
        x = self.dropout(x, key=key)

        return x

    def unembed(self, x: NamedArray):
        return hax.dot(x, self.token_embeddings, axis="embed")

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"token_embeddings": "wte.weight", "position_embeddings": "wpe.weight"}

    def resize_embeddings(self, new_size: int, key: Optional[PRNGKeyArray] = None):
        new_weights = hax.tree_util.resize_axis(self.token_embeddings, self.Vocab, new_size, key=key)
        return dataclasses.replace(self, Vocab=self.Vocab.resize(new_size), token_embeddings=new_weights)


class BackpackLMHeadModel(LmWithHfSerializationMixin, ModuleWithStateDictSerialization):
    transformer: Gpt2Transformer
    embeddings: BackpackGpt2Embeddings
    sense_net: BackpackSenses
    kq_selfattention: WeightsOnlyAttention

    @property
    def config(self):
        return self.transformer.config

    @property
    def Vocab(self) -> Axis:
        return self.embeddings.Vocab

    @property
    def Pos(self) -> Axis:
        return self.sense_net.Pos

    @staticmethod
    def init(Vocab: Axis, config: BackpackConfig, *, key):
        k_t, k_embeddings, k_attn = jrandom.split(key, 3)
        transformer = Gpt2Transformer.init(config, key=k_t)

        embeddings = BackpackGpt2Embeddings.init(
            Vocab=Vocab,
            config=config,
            key=k_embeddings,
        )
        sense_net = BackpackSenses.init(
            config=config,
            dropout_prob=config.embed_pdrop,
            key=k_embeddings,
        )
        kq_selfattention = WeightsOnlyAttention.init(
            config=config,
            key=k_attn,
        )

        return BackpackLMHeadModel(
            transformer=transformer,
            embeddings=embeddings,
            sense_net=sense_net,
            kq_selfattention=kq_selfattention,
        )

    @named_call
    def activations(
        self, input_ids: NamedArray, attn_mask: Optional[AttentionMask | NamedArray] = None, *, key=None
    ) -> NamedArray:
        k_embed, k_transformer, k_senses, k_sa = haliax.jax_utils.maybe_rng_split(key, 4)

        # Compute contextualization weights
        hidden_states = self.embeddings.embed(input_ids, key=k_embed)
        hidden_states = self.transformer(hidden_states, attn_mask, key=k_transformer)
        contextualization_weights = self.kq_selfattention(
            hidden_states, mask=attn_mask, layer_idx=self.config.num_layers, key=k_sa
        )  # (seq, seq, senses)

        ## Compute sense vectors
        sense_input_embeds = self.embeddings.embed_input_ids(input_ids)  # (seq, embed
        sense_vectors = self.sense_net.sense_embed(sense_input_embeds, key=k_senses)  # (seq, senses, embed)
        sense_vectors = sense_vectors.rename({self.Pos: self.config.KeyPos})

        ## Weight-and-sum
        hidden_states = hax.dot(
            contextualization_weights, sense_vectors, axis=self.config.KeyPos
        )  # (seq, senses, embed)
        hidden_states = hax.sum(hidden_states, axis=self.config.Senses)

        # Rescale - this is important for large num_senses
        scale = self.config.Senses.size
        hidden_states = hidden_states / scale

        return hidden_states

    def get_lm_head(self) -> hax.NamedArray:
        return self.embeddings.token_embeddings

    def resize_vocab(self, new_size: int, key: Optional[PRNGKeyArray] = None):
        new_embeddings = self.embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, embeddings=new_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {
            "transformer": "backpack.gpt2_model",
            "embeddings": "backpack.gpt2_model",
            "sense_net": "backpack.sense_network",
            "kq_selfattention": "backpack.sense_weight_net",
        }

    def update_state_dict(self, state_dict: StateDict, prefix: Optional[str] = None) -> StateDict:
        state_dict = super().update_state_dict(state_dict, prefix=prefix)
        # In levanter's implementation, we have a shared embedding matrix for both the word
        # embeddings and the sense embeddings
        state_dict[with_prefix(prefix, "backpack.word_embeddings.weight")] = state_dict[
            with_prefix(prefix, "backpack.gpt2_model.wte.weight")
        ]
        state_dict[with_prefix(prefix, "backpack.position_embeddings.weight")] = state_dict[
            with_prefix(prefix, "backpack.gpt2_model.wpe.weight")
        ]
        return state_dict



import logging
import os
from dataclasses import dataclass, field

import levanter
from levanter.data.metrics_monitor import LoggingMetricsMonitor, RichMetricsMonitor
from levanter.data.text import BatchTokenizer, LMDatasetConfig
from levanter.distributed import RayConfig
from levanter.logging import init_logging
from levanter.store.cache import build_or_load_cache
from levanter.tracker import NoopConfig, TrackerConfig


logger = logging.getLogger(__name__)


@dataclass
class RayCachedLMDatasetConfig(LMDatasetConfig, RayConfig):
    tracker: TrackerConfig = field(default_factory=NoopConfig)


@levanter.config.main()
def main(args: RayCachedLMDatasetConfig):
    """Caches two different kinds of datasets. It can cache a dataset from a list of urls, or a dataset from a hf dataset"""
    init_logging(".", "cache_dataset.log")
    args.initialize()

    tokenizer = args.the_tokenizer

    for split in ["train", "validation"]:
        print(f"Caching {split} to {args.cache_dir}.")
        # connect or start the actor
        batch_tokenizer = BatchTokenizer(tokenizer, enforce_eos=args.enforce_eos)
        split_cache_dir = os.path.join(args.cache_dir, split)  # type: ignore
        source = args.get_shard_source(split)

        if source is None:
            logger.warning(f"Skipping {split} because it is empty.")
            continue

        monitors: list = [RichMetricsMonitor(source.num_shards)]
        if not isinstance(args.tracker, NoopConfig):
            monitors.append(LoggingMetricsMonitor("preprocess/" + split, commit=True))

        cache = build_or_load_cache(
            cache_dir=split_cache_dir,
            input_shards=source,
            processor=batch_tokenizer,
            await_finished=False,
            monitors=monitors,
            split=split,
        )

        cache.await_finished()
        print(f"Finished caching {split} to {split_cache_dir}.")


if __name__ == "__main__":
    main()



import os
import tempfile

from levanter.data.sharded_datasource import (
    AudioTextUrlDataSource,
    ParquetDataSource,
    TextUrlDataSource,
    _sniff_format_for_dataset,
)
from test_utils import skip_if_no_soundlibs


def test_sniff_format_for_json():
    # this tests where some people use ".json" to mean a jsonlines file
    # and others use it to mean a json file

    with tempfile.NamedTemporaryFile(suffix=".json") as f:
        f.write(b'[{"text": "hello world"}, {"text": "hello world!"]')
        f.flush()
        assert _sniff_format_for_dataset(f.name) == ".json"

    with tempfile.NamedTemporaryFile(suffix=".json") as f:
        f.write(b'{"text": "hello world"}\n{"text": "hello world!"}\n')
        f.flush()
        assert _sniff_format_for_dataset(f.name) == ".jsonl"

    with tempfile.NamedTemporaryFile(suffix=".json") as f:
        f.write(b'{\n"ids": [1, 2, 3]\n}\n')
        f.flush()
        assert _sniff_format_for_dataset(f.name) == ".json"


def test_sniff_format_for_parquet():

    import pyarrow as pa
    import pyarrow.parquet as pq

    with tempfile.NamedTemporaryFile(suffix=".parquet") as f:
        table = pa.table({"col1": [1, 2, 3], "col2": ["a", "b", "c"]})
        pq.write_table(table, f.name)
        f.flush()

        assert _sniff_format_for_dataset(f.name) == ".parquet"


@skip_if_no_soundlibs
def test_resolve_audio_pointer():
    AudioTextUrlDataSource.resolve_audio_pointer("https://ccrma.stanford.edu/~jos/mp3/trumpet.mp3", 16_000)


def test_basic_parquet_datasource_read_row():

    import pyarrow as pa
    import pyarrow.parquet as pq

    with tempfile.NamedTemporaryFile(suffix=".parquet", delete=True) as f:
        # Create a simple dataset
        data = {"column1": ["value1", "value2", "value3"], "column2": [10, 20, 30]}
        table = pa.Table.from_pydict(data)
        pq.write_table(table, f.name)

        datasource = ParquetDataSource([os.path.abspath(f.name)])

        assert len(datasource.shard_names) == 1, "Expected only one shard"
        shard_name = datasource.shard_names[0]

        # sanity check: Read data starting from row 1
        row_data = list(datasource.open_shard_at_row(shard_name=shard_name, row=1))

        # Verify the output
        assert len(row_data) == 2  # We expect 2 rows starting from index 1
        assert row_data[0]["column1"] == "value2"
        assert row_data[0]["column2"] == 20
        assert row_data[1]["column1"] == "value3"
        assert row_data[1]["column2"] == 30


def test_text_url_data_source_parquet():
    import pyarrow as pa
    import pyarrow.parquet as pq

    with tempfile.NamedTemporaryFile(suffix=".parquet", delete=True) as f:
        data = {
            "text": ["line1", "line2", "line3", "line4", "line5", "line6"],
            "column2": [10, 20, 30, 40, 50, 60],
        }
        table = pa.Table.from_pydict(data)
        pq.write_table(table, f.name)

        datasource = TextUrlDataSource([os.path.abspath(f.name)], text_key="text")

        assert len(datasource.shard_names) == 1, "Expected only one shard"
        shard_name = datasource.shard_names[0]

        # Read data starting from row 2
        row_data = list(datasource.open_shard_at_row(shard_name=shard_name, row=2))

        # Verify the output
        expected_texts = ["line3", "line4", "line5", "line6"]
        assert len(row_data) == len(expected_texts), f"Expected {len(expected_texts)} rows starting from index 2"
        assert row_data == expected_texts, f"Expected texts {expected_texts}, got {row_data}"



import typing

import equinox as eqx
import jax.numpy as jnp
import numpy as np

import haliax as hax


Arrayish: typing.TypeAlias = hax.NamedArray | np.ndarray | jnp.ndarray


class RunningMean(eqx.Module):
    mean: Arrayish
    total: Arrayish

    @staticmethod
    def zeros_like(x: Arrayish) -> "RunningMean":
        return RunningMean(x * 0.0, x * 0.0)

    def add(self, x: Arrayish, total: Arrayish) -> "RunningMean":
        delta = x - self.mean
        # careful: total and self.total can be 0
        new_total = self.total + total
        ratio = hax.where(new_total, total / new_total, 0.0)
        new_mean = self.mean + delta * ratio
        new_total = self.total + total
        return RunningMean(new_mean, new_total)

    def __add__(self, other: "RunningMean"):
        return self.add(other.mean, other.total)

    def __str__(self):
        return f"RunningMean(mean={self.mean}, total={self.total})"



import jax
import numpy as np
import pytest

from levanter.utils.jax_utils import best_effort_sharding, create_fsdp_mesh
from test_utils import skip_if_not_enough_devices


def _assert_can_put_with_sharding(array, sharding):
    try:
        jax.device_put(array, sharding)
    except ValueError:
        # assert False, f"Could not put array with shape {array.shape} with sharding {sharding}"
        raise AssertionError(f"Could not put array with shape {array.shape} with sharding {sharding}")


@skip_if_not_enough_devices(8)
def test_best_effort_sharding():
    if len(jax.devices()) % 8 != 0:
        pytest.skip("Not enough devices")
    # 1D array, 8 devices
    array = np.arange(8)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(2, 4)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(4, 2)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(8, 1)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(1, 8)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(2, 2, 2)
    sharding = best_effort_sharding(array.shape)
    _assert_can_put_with_sharding(array, sharding)

    devices = jax.devices()[:6]

    array = array.reshape(2, 2, 2)
    sharding = best_effort_sharding(array.shape, devices=devices)
    _assert_can_put_with_sharding(array, sharding)


@pytest.mark.parametrize("fsdp_size", [1, 2, 4, 8])
def test_best_effort_sharding_with_mesh(fsdp_size):
    if fsdp_size > len(jax.devices()):
        pytest.skip("Not enough devices")
    elif len(jax.devices()) % fsdp_size != 0:
        pytest.skip("Number of devices is not a multiple of fsdp_size")

    mesh = create_fsdp_mesh(len(jax.devices()) // fsdp_size, fsdp_size, 1)

    array = np.arange(8)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(2, 4)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(4, 2)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(8, 1)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(1, 8)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)

    array = array.reshape(2, 2, 2)
    sharding = best_effort_sharding(array.shape, mesh=mesh)
    _assert_can_put_with_sharding(array, sharding)



import asyncio
import logging
from abc import ABC
from typing import Any, Callable, Dict, List, Optional, TypeVar

import ray


V = TypeVar("V")
R = TypeVar("R")

logger = logging.getLogger(__name__)

# Copilot-Adapted from:
# https://github.com/ray-project/ray/blob/1bab09bf842edee51c3778be4cfb16f8b900d764/python/ray/data/_internal/execution/operators/actor_pool_map_operator.py


def _wrap_ray_future(ray_future):
    # work around https://github.com/ray-project/ray/issues/45895#issuecomment-2165164129
    return asyncio.wrap_future(ray_future.future())


class AutoScalingActorPool:
    """Utility class to operate on a dynamically scaling pool of actors."""

    def __init__(
        self,
        create_actor_fn: Callable[[], "ray.actor.ActorHandle"],
        min_size: int = 1,
        max_size: int = 10,
    ):
        if max_size < min_size:
            raise ValueError("max_size must be greater than or equal to min_size.")
        self._create_actor_fn = create_actor_fn
        self._min_size = min_size
        self._max_size = max_size

        self._idle_actors: List[ray.actor.ActorHandle] = []
        self._busy_actors: Dict[ray.ObjectRef, ray.actor.ActorHandle] = {}
        self._pending_actors: Dict[ray.ObjectRef, ray.actor.ActorHandle] = {}

        self._actor_locations: Dict[ray.actor.ActorHandle, str] = {}
        self._tasks_waiting_for_actor: list[asyncio.Future] = []
        self._next_task_id = 0
        self._scale_down_task: Optional[asyncio.Task] = None

        self._scale_up(self._min_size)

    @property
    def num_pending_tasks(self):
        return len(self._tasks_waiting_for_actor)

    def resize_pool(self, *, min_size: Optional[int] = None, max_size: Optional[int] = None):
        old_min_size = self._min_size
        if min_size is not None:
            self._min_size = min_size
        old_max_size = self._max_size
        if max_size is not None:
            self._max_size = max_size

        if old_min_size != self._min_size or old_max_size != self._max_size:
            logger.info(f"Resizing pool to min_size: {self._min_size}, max_size: {self._max_size}")

        self._adjust_pool_size()

    def get_max_size(self):
        return self._max_size

    def get_min_size(self):
        return self._min_size

    def _scale_up(self, num_actors: int):
        if self._scale_down_task and not self._scale_down_task.done():
            self._scale_down_task.cancel()

        for _ in range(num_actors):
            try:
                actor = self._create_actor_fn()
                ready_ref = actor.get_location.remote()
                self._pending_actors[ready_ref] = actor

                async def wait_for_ready(actor, ready_ref):
                    loc = await _wrap_ray_future(ready_ref)
                    # pending -> floating
                    if ready_ref not in self._pending_actors:
                        logger.info("Actor was cancelled before it was ready.")
                        return
                    del self._pending_actors[ready_ref]
                    self._assert_is_floating(actor)
                    self._actor_locations[actor] = loc
                    self._maybe_start_pending_task(actor)  # floating -> {idle, busy}

                asyncio.ensure_future(wait_for_ready(actor, ready_ref))

            except Exception as e:
                logger.error("Failed to create actor.", exc_info=e)

    def _scale_down(self, target_num_actors: int):
        while len(self._idle_actors) + len(self._pending_actors) > target_num_actors:
            if self._pending_actors:
                actor = self._pending_actors.popitem()[1]
                # let it die through gc
                # ray.kill(actor)
            elif self._idle_actors:
                actor = self._idle_actors.pop()
                del self._actor_locations[actor]
                # let it die through gc
                # ray.kill(actor)
            else:
                break

    def _adjust_pool_size(self):
        num_pending_tasks = self.num_pending_tasks
        num_idle_actors = len(self._idle_actors)
        num_busy_actors = len(self._busy_actors)
        num_pending_actors = len(self._pending_actors)

        num_nonworking_actors = num_idle_actors + num_pending_actors
        total_actors = num_nonworking_actors + num_busy_actors

        # TODO: better autoscale logic
        if (
            num_pending_actors == 0
            and num_pending_tasks > 0
            and num_idle_actors == 0
            and total_actors < self._max_size
        ):
            logger.info(
                f"Scaling up due to {num_pending_tasks} pending tasks. Current pool size: {total_actors}. Max size:"
                f" {self._max_size}"
            )
            self._scale_up(min(self._max_size - num_busy_actors, num_pending_tasks))

        # Schedule scale down if idle
        elif num_pending_tasks == 0 and num_nonworking_actors > self._min_size:
            if self._scale_down_task is None:
                self._scale_down_task = asyncio.create_task(self._schedule_scale_down())

    async def _schedule_scale_down(self):
        try:
            await asyncio.sleep(10)
            if self.num_pending_tasks == 0:
                logger.info("Scaling down due to no pending tasks.")
                self._scale_down(self._min_size)
                self._scale_down_task = None
        except asyncio.CancelledError:
            logger.debug("Scale down task was cancelled due to new activity.")

    def _get_object_location(self, obj_ref: ray.ObjectRef) -> Optional[str]:
        """Get the location of the given object reference."""
        try:
            locs = ray.experimental.get_object_locations([obj_ref])
            nodes = locs[obj_ref]["node_ids"]
            if nodes:
                return nodes[0]
        except Exception as e:
            logger.error(f"Failed to get object location: {e}")
        return None

    def _pick_actor(self, obj_ref: Optional[ray.ObjectRef] = None) -> Optional[ray.actor.ActorHandle]:
        """Pick an actor based on locality and busyness."""
        # idle -> floating
        if not self._idle_actors:
            return None

        if obj_ref:
            preferred_loc = self._get_object_location(obj_ref)
        else:
            preferred_loc = None

        def penalty_key(actor):
            """Returns the key that should be minimized for the best actor."""
            requires_remote_fetch = self._actor_locations[actor] != preferred_loc
            return requires_remote_fetch

        actor = min(self._idle_actors, key=penalty_key)
        actor = self._idle_actors.pop(self._idle_actors.index(actor))
        return actor

    def submit(self, fn: Callable[["ray.actor.ActorHandle", V], R], value: V, obj_ref: Optional[ray.ObjectRef] = None):
        actor = self._pick_actor(obj_ref)
        if actor:
            return self._assign_task_to_actor(actor, fn, value)
        else:
            actor_future: asyncio.Future = asyncio.Future()
            self._tasks_waiting_for_actor.append(actor_future)
            f = asyncio.ensure_future(self._enqueue_pending_task(fn, obj_ref, value, actor_future))
            self._adjust_pool_size()
            return f

    def _assign_task_to_actor(self, actor, fn, value):
        # floating -> busy
        ray_future = fn(actor, value)
        self._busy_actors[ray_future] = actor
        if self._scale_down_task and not self._scale_down_task.done():
            self._scale_down_task.cancel()
        self._adjust_pool_size()

        return asyncio.ensure_future(self._set_up_actor_return_on_finished(ray_future))

    async def _enqueue_pending_task(self, fn, obj_ref, value, actor_future):
        actor = await actor_future
        return await self._assign_task_to_actor(actor, fn, value)

    def _assert_is_floating(self, actor):
        assert actor not in self._idle_actors
        assert actor not in self._busy_actors
        assert actor not in self._pending_actors

    def _maybe_start_pending_task(self, actor):
        self._assert_is_floating(actor)
        if self._tasks_waiting_for_actor:
            # floating -> busy (inside the _enqueue_pending_task coroutine)
            actor_future = self._tasks_waiting_for_actor.pop(0)
            actor_future.set_result(actor)
            assigned = True
        else:
            # floating -> idle
            self._idle_actors.append(actor)
            self._adjust_pool_size()
            assigned = False
        return assigned

    async def _set_up_actor_return_on_finished(self, ray_future):
        future = _wrap_ray_future(ray_future)
        await asyncio.wait([future])
        self._on_task_done(ray_future)
        return await future

    def _on_task_done(self, ray_future):
        actor = self._busy_actors.pop(ray_future)
        self._maybe_start_pending_task(actor)

    async def map(
        self,
        fn: Callable[["ray.actor.ActorHandle", V], Any],
        values: List[V],
        obj_refs: Optional[List[Optional[ray.ObjectRef]]] = None,
    ) -> List[Any]:
        if obj_refs is None:
            obj_refs = [None] * len(values)

        tasks = [self.submit(fn, v, obj_ref) for v, obj_ref in zip(values, obj_refs)]
        return await asyncio.gather(*tasks)

    def has_free(self):
        return bool(self._idle_actors)

    def has_free_or_pending_actors(self):
        return bool(self._idle_actors) or bool(self._pending_actors)

    def pop_idle(self):
        if self._idle_actors:
            return self._idle_actors.pop()
        return None

    def push(self, actor: "ray.actor.ActorHandle"):
        location = ray.get(actor.get_location.remote())
        self._actor_locations[actor] = location
        self._maybe_start_pending_task(actor)

    def __del__(self):
        if self._scale_down_task and not self._scale_down_task.done():
            self._scale_down_task.cancel()
        # just let ray kill the actors naturally


class PoolWorkerBase(ABC):
    def get_location(self) -> str:
        return ray.get_runtime_context().get_node_id()



import tempfile

import jax.numpy as jnp
import jmp
import numpy as np
import pytest
from chex import assert_trees_all_close, assert_trees_all_equal
from jax.random import PRNGKey

import haliax

from levanter.compat.hf_checkpoints import _convert_to_jnp
from levanter.models.attention import AttentionMask
from levanter.models.backpack import BackpackConfig, BackpackLMHeadModel
from levanter.models.gpt2 import Gpt2Config, Gpt2LMHeadModel
from levanter.utils.tree_utils import inference_mode
from test_utils import skip_if_no_torch


@skip_if_no_torch
def test_save_backpack_model_with_code():
    import torch

    converter = BackpackConfig().hf_checkpoint_converter()
    tokenizer = converter.tokenizer
    cls = converter.HFAutoModelClass()
    config = converter.HfConfigClass(
        n_embd=32,
        n_positions=512,
        n_head=8,
        n_layer=2,
        vocab_size=len(tokenizer),
        resid_pdrop=0.0,
    )

    model = cls(config)
    model.eval()

    with tempfile.TemporaryDirectory() as tmpdir:
        lev_config = converter.config_from_hf_config(config)
        model.save_pretrained(tmpdir, safe_serialization=False)  # unsafe b/c weight tying
        loaded_checkpoint = converter.load_state_dict(tmpdir)

    roundtrip_hf_config = converter.hf_config_from_config(lev_config)

    for k, v in roundtrip_hf_config.__dict__.items():
        assert getattr(roundtrip_hf_config, k) == v, f"{k} {getattr(roundtrip_hf_config, k)} != {v}"

    Vocab = converter.Vocab
    lev_model = BackpackLMHeadModel.init(Vocab, lev_config, key=PRNGKey(0))
    lev_model = haliax.state_dict.from_torch_compatible_state_dict(lev_model, loaded_checkpoint)
    lev_model = inference_mode(lev_model, True)

    with tempfile.TemporaryDirectory() as tmpdir:
        converter._save_pretrained_local(
            lev_model, tmpdir, save_tokenizer=True, save_reference_code=True, max_shard_size=1e8
        )

        new_converter = converter.replaced(reference_checkpoint=tmpdir, trust_remote_code=True)

        assert new_converter.config_from_hf_config(config) == lev_config
        loaded_model = new_converter.load_pretrained(new_converter.default_config.model_type)
        loaded_model = inference_mode(loaded_model, True)

        assert loaded_model.config == lev_model.config
        assert loaded_model.Vocab == lev_model.Vocab

        input = haliax.random.randint(PRNGKey(0), lev_model.config.Pos, 0, lev_model.Vocab.size)
        causal_mask = AttentionMask.causal()
        np.testing.assert_equal(
            np.array(lev_model(input, causal_mask, key=None).array),
            np.array(loaded_model(input, causal_mask, key=None).array),
        )

        # now double check that the pytorch model is the same
        loaded_model = cls.from_pretrained(tmpdir)
        torch_input = torch.from_numpy(np.array(input.array)).to(torch.int64).unsqueeze(0)
        loaded_model.eval()
        np.testing.assert_allclose(
            model(torch_input).logits[0].detach().numpy(),
            loaded_model(torch_input).logits[0].detach().numpy(),
            rtol=1e-3,
            atol=1e-3,
        )


@skip_if_no_torch
def test_conversion_to_jnp_bfloat16():
    import torch

    x = torch.arange(10, dtype=torch.bfloat16) / 3.14
    with pytest.raises(TypeError):
        x.cpu().numpy()

    x_jnp = _convert_to_jnp(x, None)
    assert x_jnp.dtype == jnp.bfloat16
    assert x_jnp.shape == x.shape
    assert_trees_all_close(x_jnp, jnp.arange(10, dtype=jnp.bfloat16) / 3.14)


def test_save_sharded_checkpoints():
    nano_config = Gpt2Config(hidden_dim=64, num_heads=2, num_layers=2, resid_pdrop=0.0, use_flash_attention=False)
    converter = nano_config.hf_checkpoint_converter()

    nano_model = Gpt2LMHeadModel.init(converter.Vocab, nano_config, key=PRNGKey(3))

    mp = jmp.get_policy("f32")
    nano_model = mp.cast_to_param(nano_model)

    with tempfile.TemporaryDirectory() as tmpdir:
        converter.save_pretrained(nano_model, tmpdir, max_shard_size=1024)

        # make sure we saved a few different files
        import glob

        assert len(glob.glob(tmpdir + "/*.safetensors")) > 1

        loaded_model = converter.load_pretrained(
            Gpt2LMHeadModel, ref=tmpdir, config=nano_model.config, dtype=mp.param_dtype
        )

        assert loaded_model.config == nano_model.config
        assert loaded_model.Vocab == nano_model.Vocab

        assert_trees_all_equal(
            nano_model,
            loaded_model,
        )



import os
import tempfile

import jax
import pytest

import haliax

import levanter.main.viz_logprobs as viz_logprobs
import tiny_test_corpus
from levanter.checkpoint import save_checkpoint
from levanter.distributed import RayConfig
from levanter.models.gpt2 import Gpt2Config, Gpt2LMHeadModel
from levanter.tracker.wandb import WandbConfig


@pytest.mark.entry
def test_viz_lm():
    # just testing if eval_lm has a pulse
    # save a checkpoint
    model_config = Gpt2Config(
        num_layers=2,
        num_heads=2,
        hidden_dim=32,
        seq_len=64,
        use_flash_attention=True,
    )

    with tempfile.TemporaryDirectory() as f:
        try:
            data_config, _ = tiny_test_corpus.construct_small_data_cache(f)
            tok = data_config.the_tokenizer
            Vocab = haliax.Axis("vocab", len(tok))
            model = Gpt2LMHeadModel.init(Vocab, model_config, key=jax.random.PRNGKey(0))

            save_checkpoint({"model": model}, 0, f"{f}/ckpt")

            config = viz_logprobs.VizGpt2Config(
                data=data_config,
                model=model_config,
                trainer=viz_logprobs.TrainerConfig(
                    per_device_eval_parallelism=len(jax.devices()),
                    max_eval_batches=1,
                    wandb=WandbConfig(mode="disabled"),
                    require_accelerator=False,
                    ray=RayConfig(auto_start_cluster=False),
                ),
                checkpoint_path=f"{f}/ckpt",
                num_docs=len(jax.devices()),
                path=f"{f}/viz",
            )
            viz_logprobs.main(config)
        finally:
            try:
                os.unlink("wandb")
            except Exception:
                pass



import logging
from dataclasses import dataclass, field
from typing import Union

import equinox as eqx
import jax.random as jrandom

from haliax import Axis
from haliax.partitioning import named_jit, round_axis_for_partitioning

import levanter
from levanter.compat.hf_checkpoints import HFCompatConfig
from levanter.data.text import CausalLmDataset, LMMixtureDatasetConfig
from levanter.doremi import DoReMiConfig, estimate_mixture_weights
from levanter.models.gpt2 import Gpt2Config
from levanter.models.lm_model import LmConfig, LmHeadModel, compute_next_token_loss
from levanter.optim import AdamConfig, OptimizerConfig
from levanter.trainer import TrainerConfig
from levanter.utils.tree_utils import inference_mode


logger = logging.getLogger(__name__)


@dataclass
class TrainLmConfig:
    ref_model_path: str
    ref_model_from_hf: bool = False

    data: LMMixtureDatasetConfig = field(default_factory=LMMixtureDatasetConfig)
    trainer: TrainerConfig = field(default_factory=TrainerConfig)
    model: LmConfig = field(default_factory=Gpt2Config)
    optimizer: OptimizerConfig = field(default_factory=AdamConfig)
    doremi: DoReMiConfig = field(default_factory=DoReMiConfig)

    # config related to continued pretraining
    initialize_from_hf: Union[bool, str] = False
    """if provided, this will override the model config in the config. if true, use the default hf checkpoint for this model class"""
    use_hf_model_config: bool = False  # if true, replace the model config with the hf config from the checkpoint

    # TODO: atm we don't support loading from a checkpoint that has a different tokenizer. this is a bit annoying
    # TODO: atm you have to at least specify a levanter model config with the same type as the hf checkpoint


def main(config: TrainLmConfig):
    levanter.initialize(config)

    tokenizer = config.data.the_tokenizer

    # this is some unpleasant code to allow us to initialize from a hf checkpoint. If this is your first read through,
    # I recommend skipping it for now
    if config.initialize_from_hf:
        if config.trainer.initialize_from is not None:
            raise ValueError("Cannot specify both initialize_from_hf and initialize_from")

        assert isinstance(config.model, HFCompatConfig)
        converter = config.model.hf_checkpoint_converter()
        if hasattr(tokenizer, "vocab") and tokenizer.vocab != converter.tokenizer.vocab:
            logger.warning("The tokenizers appear to be different. You may want to check this.")

        if isinstance(config.initialize_from_hf, str):
            converter = converter.replaced(reference_checkpoint=config.initialize_from_hf, tokenizer=tokenizer)
        else:
            converter = converter.replaced(tokenizer=tokenizer)

        if config.use_hf_model_config:
            # TODO: log diff of old and new config
            # NB: gross mutability
            config.model = converter.config_from_hf_config(converter.default_hf_config)
    elif isinstance(config.model, HFCompatConfig):
        converter = config.model.hf_checkpoint_converter()
        converter = converter.replaced(tokenizer=tokenizer)
    else:
        converter = None

    optimizer = config.optimizer.build(config.trainer.num_train_steps)

    parameter_axis_mapping = config.trainer.parameter_axis_mapping

    loss_function = compute_next_token_loss

    with config.trainer.device_mesh:
        vocab_size = len(tokenizer)
        Vocab = round_axis_for_partitioning(Axis("vocab", vocab_size), parameter_axis_mapping)
        if vocab_size != Vocab.size:
            logger.info(f"Rounding vocab size from {vocab_size} to {Vocab.size} for partitioning")

        # initialize the ref model
        if config.ref_model_from_hf:
            assert converter is not None
            ref_model = converter.load_pretrained(config.model.model_type, dtype=config.trainer.mp.compute_dtype)
        else:
            ref_model_shape = eqx.filter_eval_shape(config.model.build, Vocab, key=jrandom.PRNGKey(0))
            ref_model = levanter.checkpoint.load_checkpoint(
                ref_model_shape, config.ref_model_path, axis_mapping=parameter_axis_mapping, subpath="model"
            )

        ref_model = inference_mode(ref_model, True)
        assert isinstance(ref_model, LmHeadModel)

        training_key, model_key = jrandom.split(jrandom.PRNGKey(config.trainer.seed), 2)

        @named_jit(axis_resources=parameter_axis_mapping)
        def init_proxy_model():
            return config.model.build(Vocab, key=model_key)

        proxy_model = init_proxy_model()

        train_datasets = config.data.training_sets(ref_model.Pos.size)
        valid_datasets = config.data.validation_sets(ref_model.Pos.size)

        causal_train_datasets = {
            k: CausalLmDataset(v, config.model.Pos, config.model.KeyPos, ignore_index=config.data.ignore_token_id)
            for k, v in train_datasets.items()
        }
        valid_datasets = {
            k: CausalLmDataset(v, config.model.Pos, config.model.KeyPos, ignore_index=config.data.ignore_token_id)
            for k, v in valid_datasets.items()
        }

        mixture_weights = estimate_mixture_weights(
            loss_function,
            proxy_model,
            ref=ref_model,
            data_sources=causal_train_datasets,
            trainer_config=config.trainer,
            optimizer=optimizer,
            domain_weight_step_size=config.doremi.domain_weight_step_size,
            sampling_weights=config.doremi.sampling_weights,
            validation_sets=valid_datasets,
            key=training_key,
        )

        print(mixture_weights)

        # dump to a yaml file
        weights_path = "mixture_weights.yaml"
        with open(weights_path, "w") as f:
            import yaml

            yaml.dump(mixture_weights, f)

        # log as an artifact
        levanter.tracker.current_tracker().log_artifact(weights_path, name="mixture_weights.yaml")


if __name__ == "__main__":
    levanter.config.main(main)()



from typing import Iterable, Iterator, List, TypeVar


T = TypeVar("T")


def batched(iterable: Iterable[T], batch_size: int) -> Iterator[List[T]]:
    """Yields batches of the given size from the given iterable."""
    batch = []
    for item in iterable:
        batch.append(item)
        if len(batch) == batch_size:
            yield batch
            batch = []

    if len(batch) > 0:
        yield batch
